"""
Some functions to compute Multi-Agent Systems with NuSMV.
"""

from pynusmv_lower_interface.nusmv.compile import compile as nscompile
from pynusmv_lower_interface.nusmv.node import node as nsnode
from pynusmv_lower_interface.nusmv.parser import parser as nsparser
from pynusmv_lower_interface.nusmv.set import set as nsset
from pynusmv_lower_interface.nusmv.utils import utils as nsutils
from pynusmv_lower_interface.nusmv.compile.symb_table import symb_table as nssymb_table

from pynusmv.exception import NuSMVNoReadModelError
from pynusmv.parser import parse_next_expression

from pynusmv.glob import (load_from_file, load,
                               flatten_hierarchy as _flatten_hierarchy,
                               symb_table, bdd_encoding,
                               prop_database as _prop_database,
                               compute_model as _compute_model)
                               
from .mas import MAS, Group

import itertools


# The current multi-agent system
__mas = None

def reset_globals():
    """
    Reset mas.glob related global variables.
    
    Must be called whenever (and before) pynusmv.init.init.deinit_nusmv
    is called.    
    """
    global __mas
    __mas = None


def _get_instances_args_for_module(modtree):
    """
    Return a dictionary of instance name -> list of instance arguments pairs,
    with instances of modules in module modtree.
    
    modtree is a part of the AST of the SMV model. modtree.type = MODULE
    """
    # MODULE(MODTYPE(...), declarationlist)
    varlist = {}
    
    declarations = nsnode.cdr(modtree)
    while declarations is not None:
        decl = nsnode.car(declarations)
        
        if decl.type == nsparser.VAR:
            
            decl = nsnode.car(decl)
            while decl is not None:
                var = nsnode.car(decl)
                # var_id : type => COLON(ATOM, type)
                if nsnode.cdr(var).type == nsparser.MODTYPE:
                    varid = nsnode.sprint_node(nsnode.car(var))
                    if varid in varlist:
                        pass # TODO Variable already defined
                    else:
                        # Compute args list
                        argslist = []
                        args = nsnode.cdr(nsnode.cdr(var))
                        while args is not None:
                            arg = nsnode.car(args)
                            argslist.append(arg)
                            args = nsnode.cdr(args)
                        varlist[varid] = argslist
                decl = nsnode.cdr(decl)
        
        declarations = nsnode.cdr(declarations)
        
    return varlist
    
    
def _flatten_and_filter_variable_args(arguments):
    """
    Return a new dictionary instance name -> list of vars where
    all vars belong to arguments (under the correct instance name)
    all vars are VAR types
    all vars are flattened.
    
    arguments -- a dictionary instance name -> list of module arguments
    """
    result = {}
    st = symb_table()
    for instance in arguments:
        result[instance] = []
        for argument in arguments[instance]:
            arg, err = nscompile.FlattenSexp(st._ptr, argument, None)
            if not err and nssymb_table.SymbTable_is_symbol_var(st._ptr, arg):
                result[instance].append(arg)
    return result
    
    
def _get_variables_by_instances(agents):
    """
    Return a dictionary of instance->list of variables
    """
    st = symb_table()
    flatHierarchy = nscompile.cvar.mainFlatHierarchy
    
    # Populate variables with instances
    variables = {}
    for agent in agents:
        variables[agent] = []
    
    varset = nscompile.FlatHierarchy_get_vars(flatHierarchy)
    varlist = nsset.Set_Set2List(varset)
    
    ite = nsutils.NodeList_get_first_iter(varlist)
    while not nsutils.ListIter_is_end(ite):
        var = nsutils.NodeList_get_elem_at(varlist, ite)
        varname = nsnode.sprint_node(var)
        isVar = nssymb_table.SymbTable_is_symbol_state_var(st._ptr, var)
        if isVar:
            # Put the var in the variables dictionary, under the right instance
            topcontext = varname.partition(".")[0]
            if topcontext in variables:
                variables[topcontext].append(var)                    
        ite = nsutils.ListIter_get_next(ite)
        
    return variables
    

def _get_input_vars_by_instances(agents):
    """
    Return a dictionary of instance->list of input variables
    """
    st = symb_table()
    flatHierarchy = nscompile.cvar.mainFlatHierarchy
    
    # Populate variables with instances
    variables = {}
    for agent in agents:
        variables[agent] = []
    
    varset = nscompile.FlatHierarchy_get_vars(flatHierarchy)
    varlist = nsset.Set_Set2List(varset)
    
    ite = nsutils.NodeList_get_first_iter(varlist)
    while not nsutils.ListIter_is_end(ite):
        var = nsutils.NodeList_get_elem_at(varlist, ite)
        varname = nsnode.sprint_node(var)
        isVar = nssymb_table.SymbTable_is_symbol_input_var(st._ptr, var)
        if isVar:
            # Put the var in the variables dictionary, under the right instance
            topcontext = varname.partition(".")[0]
            if topcontext in variables:
                variables[topcontext].append(var)                    
        ite = nsutils.ListIter_get_next(ite)
        
    return variables
    
        
        
def _get_epistemic_trans(variable):
    """Return a new TRANS next(variable) = variable"""
    
    transexpr = "next({0}) = {0}".format(variable)
    return parse_next_expression(transexpr)
    

def mas(agents=None, initial_ordering=None):
    """
    Return (and compute if needed) the multi-agent system represented by
    the currently read SMV model.
    
    If agents is not None, the set of agents (and groups) of the MAS is
    determined by agents.
    
    Otherwise, every top-level module instantiation is considered an agent
    where
        - her actions are the inputs variables prefixed by her name;
        - her observable variables are composed of
            * the state variables of the system prefixed by her name;
            * the state variables provided as an argument to the module
              instantiation.
    
    If initial_ordering is not None, it must be the path to a variables
    ordering file. It is used as the initial ordering for variables of the
    model.
    
    Note: if the MAS is already computed, agents and initial_ordering arguments
    have no effect.
    
    agents -- a set of agents.
    """    
    global __mas
    if __mas is None:
        # Check cmps
        if not nscompile.cmp_struct_get_read_model(nscompile.cvar.cmps):
            raise NuSMVNoReadModelError("Cannot build MAS; no read file.")
        
        if agents is None:
            # Get agents names
            tree = nsparser.cvar.parsed_tree
            main = None
            while tree is not None:
                module = nsnode.car(tree)
                if (nsnode.sprint_node(nsnode.car(nsnode.car(module))) ==
                    "main"):
                    main = module            
                tree = nsnode.cdr(tree)
            if main is None:
                print("[ERROR] No main module.")
                return # TODO Error, cannot find main module
            arguments = _get_instances_args_for_module(main)
            # arguments is a dict instancename(str)->listofargs(node)
            agents = arguments.keys()
            
            # Compute the model
            _compute_model(variables_ordering=initial_ordering)
            
            st = symb_table()
            
            # Flatten arguments and filter on variables
            argvars = _flatten_and_filter_variable_args(arguments)
            
            # Get agents observable variables (locals + module parameters)
            localvars = _get_variables_by_instances(agents)
            #localvars is a dict instancename(str)->listofvars(node)
            inputvars = _get_input_vars_by_instances(agents)
            
            # Merge instance variable arguments and local variables
            variables = {key: ((key in argvars and argvars[key] or []) + 
                               (key in localvars and localvars[key] or []))
                         for key in
                         list(argvars.keys())+list(localvars.keys())}
            
            # Compute epistemic relation
            singletrans = {}
            for agent in variables:
                transexpr = None
                for var in variables[agent]:
                    var = nsnode.sprint_node(var)
                    transexpr = nsnode.find_node(nsparser.AND,                                                       
                                                 _get_epistemic_trans(var),
                                                 transexpr)
                singletrans[agent] = transexpr           
            
            # Process variables to get strings instead of nodes
            observedvars = {ag: {nsnode.sprint_node(v) for v in variables[ag]}
                            for ag in variables.keys()}
            inputvars = {ag: {nsnode.sprint_node(v)
                              for v in inputvars[ag]}
                         for ag in inputvars.keys()}
            groups = None
        
        else:
            _compute_model(variables_ordering=initial_ordering)
            # observedvars: a dictionary of agent name -> set of observed vars
            observedvars = {str(agent.name): [str(var)
                                              for var in agent.observables]
                            for agent in agents}
            # inputsvars: a dictionary of agent name -> set of inputs vars
            inputvars = {str(agent.name): [str(ivar)
                                           for ivar in agent.actions]
                         for agent in agents}
            # groups:
            # a dictionary of group name -> names of agents of the group
            groups = {str(group.name): [str(agent.name)
                                        for agent in group.agents]
                      for group in agents if isinstance(group, Group)}
            # singletrans: a dictionary of agent name -> epistemic transition
            singletrans = {}
            for agent in agents:
                name = str(agent.name)
                transexpr = None
                for var in observedvars[name]:
                    transexpr = nsnode.find_node(nsparser.AND,
                                                 _get_epistemic_trans(var),
                                                 transexpr)
                singletrans[name] = transexpr
            
        
        # Create the MAS
        fsm = _prop_database().master.bddFsm
        __mas = MAS(fsm._ptr, observedvars, inputvars, singletrans,
                    groups=groups, freeit=False)
        
    return __mas


import unittest

from pynusmv.dd import BDD
from pynusmv.init import init_nusmv, deinit_nusmv
from pynusmv.mc import eval_simple_expression

from pynusmv_lower_interface.nusmv.dd import dd as nsdd

from pynusmv_tools.mas import glob

from pynusmv_tools.atlkPO.eval import split, cex_si, nfair_gamma_si, nfair_gamma
from pynusmv_tools.atlkPO.evalPartial import reach, split_reach, split as psplit

from pynusmv.utils import fixpoint as fp


class TestEval(unittest.TestCase):
    
    def setUp(self):
        init_nusmv()
    
    def tearDown(self):
        glob.reset_globals()
        deinit_nusmv()
    
    
    def little(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/little.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def tree(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/tree.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def collapsed_tree(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/collapsed-tree.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def cardgame(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/cardgame.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def cardgame_fair(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/cardgame-fair.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def cardgame_post_fair(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/cardgame-post-fair.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def transmission_post_fair(self):
        glob.load_from_file(
                        "tests/pynusmv_tools/atlkPO/models/transmission-post-fair.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def trans2_fair(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/2-transmission-fair.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def nfair_model(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/nfair.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def show_si(self, fsm, bdd):
        if bdd.isnot_false():
            sis = fsm.pick_all_states_inputs(bdd)
            print("SI count:", len(sis))
            for si in sis:
                print(si.get_str_values())
            
    def show_s(self, fsm, bdd):
        if bdd.isnot_false():
            ss = fsm.pick_all_states(bdd)
            print("S count:", len(ss))
            for s in ss:
                print(s.get_str_values())
            
    def show_i(self, fsm, bdd):
        if bdd.isnot_false():
            iss = fsm.pick_all_inputs(bdd)
            print("I count:", len(iss))
            for i in iss:
                print(i.get_str_values())


    def test_split(self):
        fsm = self.cardgame()
        
        s0 = eval_simple_expression(fsm, "step = 0")
        s1 = eval_simple_expression(fsm, "step = 1")
        s2 = eval_simple_expression(fsm, "step = 2")
        pa = eval_simple_expression(fsm, "pcard = Ac")
        pk = eval_simple_expression(fsm, "pcard = K")
        pq = eval_simple_expression(fsm, "pcard = Q")
        da = eval_simple_expression(fsm, "dcard = Ac")
        dk = eval_simple_expression(fsm, "dcard = K")
        dq = eval_simple_expression(fsm, "dcard = Q")
        dda = eval_simple_expression(fsm, "ddcard = Ac")
        ddk = eval_simple_expression(fsm, "ddcard = K")
        ddq = eval_simple_expression(fsm, "ddcard = Q")
        win = eval_simple_expression(fsm, "win")
        lose = eval_simple_expression(fsm, "lose")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        ngamma_cube = (fsm.bddEnc.inputsCube -
                       fsm.inputs_cube_for_agents({"player"}))
        
        strats = split(fsm, fsm.protocol({"player"}), {"player"})
        
        self.assertEqual(len(strats), 8)
    
    
    def test_cex_si(self):
        fsm = self.cardgame()
        
        s0 = eval_simple_expression(fsm, "step = 0")
        s1 = eval_simple_expression(fsm, "step = 1")
        s2 = eval_simple_expression(fsm, "step = 2")
        
        pa = eval_simple_expression(fsm, "pcard = Ac")
        pk = eval_simple_expression(fsm, "pcard = K")
        pq = eval_simple_expression(fsm, "pcard = Q")
        
        da = eval_simple_expression(fsm, "dcard = Ac")
        dk = eval_simple_expression(fsm, "dcard = K")
        dq = eval_simple_expression(fsm, "dcard = Q")
        
        dda = eval_simple_expression(fsm, "ddcard = Ac")
        ddk = eval_simple_expression(fsm, "ddcard = K")
        ddq = eval_simple_expression(fsm, "ddcard = Q")
        
        pan = eval_simple_expression(fsm, "player.action = none")
        pak = eval_simple_expression(fsm, "player.action = keep")
        pas = eval_simple_expression(fsm, "player.action = swap")
        
        dan = eval_simple_expression(fsm, "dealer.action = none")
        daak = eval_simple_expression(fsm, "dealer.action = dealAK")
        
        win = eval_simple_expression(fsm, "win")
        lose = eval_simple_expression(fsm, "lose")
        
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        dealercube = fsm.inputs_cube_for_agents({'dealer'})
        playercube = fsm.inputs_cube_for_agents({'player'})
        
        
        self.assertTrue(pk & dq & s2 & pan & fsm.reachable_states & fsm.bddEnc.statesInputsMask <= cex_si(fsm, {'player'}, fsm.reachable_states & win))
        self.assertTrue(pk & dq & s2 & pan & fsm.reachable_states & fsm.bddEnc.statesInputsMask <= cex_si(fsm, {'player'}, fsm.reachable_states & win, pk & dq & pan))
        self.assertTrue(pk & dq & s1 & pak & fsm.reachable_states & fsm.bddEnc.statesInputsMask <= cex_si(fsm, {'player'}, fsm.reachable_states & win, pak))
        self.assertFalse(pk & dq & s1 & pas & fsm.reachable_states & fsm.bddEnc.statesInputsMask <= cex_si(fsm, {'player'}, fsm.reachable_states & win, pak))
        self.assertTrue(pa & dq & s1 & pas & fsm.reachable_states & fsm.bddEnc.statesInputsMask <= cex_si(fsm, {'player'}, fsm.reachable_states & win))
        self.assertFalse(pa & dq & s1 & pas & fsm.reachable_states & fsm.bddEnc.statesInputsMask <= cex_si(fsm, {'player'}, fsm.reachable_states & win, pak))
        
        strat = (s0 & daak & fsm.protocol({'dealer'})) | ((s2) & fsm.protocol({'dealer'}))
        self.assertTrue(nfair_gamma_si(fsm, {'dealer'}, strat).is_false())
        self.assertEqual(pa, nfair_gamma_si(fsm, {'dealer'}, strat) | pa)
        cexsi = cex_si(fsm, {'dealer'}, fsm.reachable_states & pa, strat)
        self.assertTrue(fsm.init <= cexsi.forsome(fsm.bddEnc.inputsCube))
        
        
    
    def test_nfair_gamma_si(self):
        fsm = self.cardgame_post_fair()
        
        s0 = eval_simple_expression(fsm, "step = 0")
        s1 = eval_simple_expression(fsm, "step = 1")
        s2 = eval_simple_expression(fsm, "step = 2")
        
        pa = eval_simple_expression(fsm, "pcard = Ac")
        pk = eval_simple_expression(fsm, "pcard = K")
        pq = eval_simple_expression(fsm, "pcard = Q")
        
        da = eval_simple_expression(fsm, "dcard = Ac")
        dk = eval_simple_expression(fsm, "dcard = K")
        dq = eval_simple_expression(fsm, "dcard = Q")
        
        dda = eval_simple_expression(fsm, "ddcard = Ac")
        ddk = eval_simple_expression(fsm, "ddcard = K")
        ddq = eval_simple_expression(fsm, "ddcard = Q")
        
        pan = eval_simple_expression(fsm, "player.action = none")
        pak = eval_simple_expression(fsm, "player.action = keep")
        pas = eval_simple_expression(fsm, "player.action = swap")
        
        dan = eval_simple_expression(fsm, "dealer.action = none")
        
        win = eval_simple_expression(fsm, "win")
        lose = eval_simple_expression(fsm, "lose")
        
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        
        agents = {'dealer'}
        strats = split(fsm, fsm.protocol(agents), agents)
        strat = strats.pop()
        nf = ~fsm.fairness_constraints[0] & fsm.bddEnc.statesInputsMask
        
        self.assertEqual(nf & fsm.pre_strat_si(BDD.true(fsm.bddEnc.DDmanager), agents, strat), nf & fsm.pre_strat_si(BDD.true(fsm.bddEnc.DDmanager), agents, strat) & fsm.bddEnc.statesInputsMask)
        
        nfp = nfair_gamma_si(fsm, {'player'})
        nfd = nfair_gamma_si(fsm, {'dealer'})
        
        self.assertTrue(nfp.is_false())
        self.assertTrue(fsm.protocol({'dealer'}) <= nfd)
        
        
    def test_nfair_gamma_si_trans2_fair(self):
        fsm = self.trans2_fair()
        
        #print("nfair_[transmitter] SI without strategies")
        #self.show_si(fsm, nfair_gamma_si(fsm, {'transmitter'}) &
        #                  fsm.protocol({'transmitter','sender'}))
        
        #print("nfair_[sender] SI without strategies")
        #self.show_si(fsm, nfair_gamma_si(fsm, {'sender'}))
        
        strats = split(fsm, fsm.protocol({'transmitter'}), {'transmitter'})
        strat = strats.pop()
        #print("nfair_[transmitter] SI in a uniform strategy")
        #self.show_si(fsm, nfair_gamma_si(fsm, {'transmitter'}, strat))
        
        # TODO Write a test
        
        
    def test_nfair_gamma_si_transmission_post_fair(self):
        fsm = self.transmission_post_fair()
        
        transmit = eval_simple_expression(fsm, "transmitter.action = transmit")
        block = eval_simple_expression(fsm, "transmitter.action = block")
        wait = eval_simple_expression(fsm, "sender.action = wait")
        send = eval_simple_expression(fsm, "sender.action = send")
        
        received = eval_simple_expression(fsm, "received")
        sent = eval_simple_expression(fsm, "sent")
        waited = eval_simple_expression(fsm, "waited")
        transmitted = eval_simple_expression(fsm, "transmitted")
        blocked = eval_simple_expression(fsm, "blocked")
        
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        nfgsit = nfair_gamma_si(fsm, {'transmitter'})
        self.assertTrue(fsm.protocol({'transmitter'}) <= nfgsit)
        
        nfgsis = nfair_gamma_si(fsm, {'sender'})
        self.assertEqual(false, nfgsis)
        
    
    def test_nfair_gamma(self):
        fsm = self.transmission_post_fair()
        
        transmit = eval_simple_expression(fsm, "transmitter.action = transmit")
        false = BDD.false(fsm.bddEnc.DDmanager)
        
        self.assertTrue(fsm.reachable_states <= nfair_gamma(fsm, {'transmitter'}))
        
        self.assertEqual(false, nfair_gamma(fsm, {'sender'}))
        
        strats = split(fsm, fsm.protocol({'transmitter'}), {'transmitter'})
        for strat in strats:
            if (strat & transmit).isnot_false():
                self.assertTrue(nfair_gamma(fsm, {'transmitter'}, strat).is_false())
            else:
                self.assertTrue(fsm.reachable_states <= nfair_gamma(fsm, {'transmitter'}, strat))
                
                
    def test_nfair_gamma_cardgame_post_fair(self):
        fsm = self.cardgame_post_fair()
        
        s0 = eval_simple_expression(fsm, "step = 0")
        s1 = eval_simple_expression(fsm, "step = 1")
        s2 = eval_simple_expression(fsm, "step = 2")
        
        pa = eval_simple_expression(fsm, "pcard = Ac")
        pk = eval_simple_expression(fsm, "pcard = K")
        pq = eval_simple_expression(fsm, "pcard = Q")
        
        da = eval_simple_expression(fsm, "dcard = Ac")
        dk = eval_simple_expression(fsm, "dcard = K")
        dq = eval_simple_expression(fsm, "dcard = Q")
        
        dda = eval_simple_expression(fsm, "ddcard = Ac")
        ddk = eval_simple_expression(fsm, "ddcard = K")
        ddq = eval_simple_expression(fsm, "ddcard = Q")
        
        pan = eval_simple_expression(fsm, "player.action = none")
        pak = eval_simple_expression(fsm, "player.action = keep")
        pas = eval_simple_expression(fsm, "player.action = swap")
        
        dan = eval_simple_expression(fsm, "dealer.action = none")
        
        win = eval_simple_expression(fsm, "win")
        lose = eval_simple_expression(fsm, "lose")
        
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        self.assertEqual(false, nfair_gamma(fsm, {'player'}))
        
        self.assertTrue(fsm.reachable_states & fsm.bddEnc.statesInputsMask <= nfair_gamma(fsm, {'dealer'}))
        
        strats = split(fsm, fsm.protocol({'dealer'}), {'dealer'})
        for strat in strats:
            self.assertTrue(fsm.reachable_states & fsm.bddEnc.statesInputsMask <= nfair_gamma(fsm, {'dealer'}, strat))
            
            
    def test_nfair_si_nfair_model(self):
        fsm = self.nfair_model()
        
        s0 = eval_simple_expression(fsm, "state = s0")
        s1 = eval_simple_expression(fsm, "state = s1")
        s2 = eval_simple_expression(fsm, "state = s2")
        
        a0 = eval_simple_expression(fsm, "a.a = 0")
        a1 = eval_simple_expression(fsm, "a.a = 1")
        a2 = eval_simple_expression(fsm, "a.a = 2")
        
        b0 = eval_simple_expression(fsm, "b.a = 0")
        b1 = eval_simple_expression(fsm, "b.a = 1")
        
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        nfgsia = nfair_gamma_si(fsm, {'a'})
        nfgsib  = nfair_gamma_si(fsm, {'b'})
        
        self.assertTrue(s0 & a1 & fsm.protocol({'a'}) <= nfgsia)
        self.assertTrue(s1 & a0 & fsm.protocol({'a'}) <= nfgsia)
        self.assertEqual((s0 & a1 & fsm.protocol({'a'})) |
                         (s1 & a0 & fsm.protocol({'a'})), nfgsia)
                         
        self.assertEqual(s1 & b0 & fsm.protocol({'b'}), nfgsib)
        
        
    def test_splitreach_tree(self):
        fsm = self.tree()
        agents = {"a"}
        
        s0 = eval_simple_expression(fsm, "a.state = 0")
        s1 = eval_simple_expression(fsm, "a.state = 1")
        s2 = eval_simple_expression(fsm, "a.state = 2")
        s3 = eval_simple_expression(fsm, "a.state = 3")
        s4 = eval_simple_expression(fsm, "a.state = 4")
        s5 = eval_simple_expression(fsm, "a.state = 5")
        s6 = eval_simple_expression(fsm, "a.state = 6")
        
        aa = eval_simple_expression(fsm, "a.action = a")
        ab = eval_simple_expression(fsm, "a.action = b")
        
        splitted_init = psplit(fsm, fsm.init & fsm.protocol(agents), agents)
        strats = {strat for pustrat in splitted_init
                        for strat in split_reach(fsm, agents, pustrat)}
        self.assertEqual(len(strats), 4)
        self.assertTrue(((s0 & aa) | (s1 & aa) | (s3 & aa))
                         & fsm.bddEnc.statesMask in strats)
        self.assertTrue(((s0 & aa) | (s1 & ab) | (s4 & aa))
                        & fsm.bddEnc.statesMask in strats)
        self.assertTrue(((s0 & ab) | (s2 & aa) | (s5 & aa))
                        & fsm.bddEnc.statesMask in strats)
        self.assertTrue(((s0 & ab) | (s2 & ab) | (s6 & aa))
                        & fsm.bddEnc.statesMask in strats)
                        
                        
    def test_splitreach_collapsed_tree(self):
        fsm = self.collapsed_tree()
        agents = {"a"}
        
        sa0 = eval_simple_expression(fsm, "a.state = 0")
        sa1 = eval_simple_expression(fsm, "a.state = 1")
        
        s0 = eval_simple_expression(fsm, "s = 0")
        s1 = eval_simple_expression(fsm, "s = 1")
        s2 = eval_simple_expression(fsm, "s = 2")
        
        aa = eval_simple_expression(fsm, "a.action = a")
        ab = eval_simple_expression(fsm, "a.action = b")
        
        splitted_init = psplit(fsm, fsm.init & fsm.protocol(agents), agents)
        strats = {strat for pustrat in splitted_init
                        for strat in split_reach(fsm, agents, pustrat)}
        self.assertEqual(len(strats), 2)
        
        
    def test_reach_cardgame(self):
        fsm = self.cardgame()
        self.assertEqual(fsm.reachable_states, reach(fsm, fsm.init))


#! /usr/local/bin/python3
"""
This module implements a simple verification tool able to perform a
diagnosability test as explained in::

    Formal verification of diagnosability via symbolic model checking.
    Pecheur, C., Cimatti, A., & Cimatti, R. (2002, July).
    In Workshop on Model Checking and Artificial Intelligence (MoChArt-2002),
    Lyon, France.

Beyond the verification proper, this tool provides an illustrative example to
show how the SAT-BMC addition to PyNuSMV may be used in order to perform
verification scenarios which were initially not present in NuSMV.
"""
import sys
import argparse
from re                    import fullmatch
from functools             import reduce

from pynusmv.init          import init_nusmv
from pynusmv.glob          import load, master_bool_sexp_fsm
from pynusmv.parser        import parse_simple_expression,     \
                                  parse_ltl_spec
from pynusmv.node          import Node

from pynusmv.bmc.glob      import BmcSupport, master_be_fsm
from pynusmv.be.expression import Be
from pynusmv.bmc.utils     import BmcModel,                    \
                                  make_nnf_boolean_wff,        \
                                  booleanize
from pynusmv.bmc.ltlspec   import bounded_semantics_at_offset

from pynusmv.sat           import SatSolverFactory,            \
                                  Polarity,                    \
                                  SatSolverResult


def arguments():
    """
    Thus function takes care of setting up the machinery to parse the command
    line arguments and return an object containing the parsed args. It moreover
    takes care of displaying an error message / help message whenever needed.

    :return: an object containing the parsed command-line arguments.
    """
    args = argparse.ArgumentParser(description="""
        This tool implements a diagnosability test verifier with SAT BMC.
    """)
    args.add_argument("model",
                    help="The model to load")

    args.add_argument("-q", "--quiet", action="store_true",
                      help="Quiet mode: disable the verbose greeting outputs")

    args.add_argument("-s", "--spec",
                      help="The property whose diagnosability needs to be verified"+\
                           "diagnosability condition *NEED* to be separated by a semicolon")

    args.add_argument("-oi","--observable-inputs", action="store_true",
                      help="Makes all input variables (IVAR) observable")

    args.add_argument("-of","--observable-file",
                      help="A file from where to load the configuration of the"+\
                      " variables that are observable. In the file, all items "+\
                      "are separated either by a semicolon or a new line. All "+\
                      "items are considered to be regexes that match the name "+\
                      "of some of the symbols (which need to be observable)")

    args.add_argument("-or","--observable-regex", action="append",
                      default=[],
                      help="Adds a regex to the list of patterns that are used"+\
                      " in order to determine which variables must be"+\
                      " considered observable for the diagnosability test(s)")

    args.add_argument("-o", "--observable", action="append",
                      default=[],
                      help="Make the symbol passed as argument visible in the " +\
                      "current diagnosability test. This option can be repeated"+\
                      "at will to have multiple visible variables.")
    args.add_argument("-k", "--bound",
                      type=int,
                      default=10,
                      help="The problem bound (max number of steps in a trace)")

    # Definition of the Diagnosability context (\theta, \Sigma_{12})
    args.add_argument("-i", "--initial-condition", default="TRUE",
                      help="An initial condition which is used to constrain the"+\
                      " possible initial belief states. This option should be " +\
                      "used in order to force the value of some non-observable "+\
                      "variable (with a non deterministic initial value) in the"+\
                      " initial state in order to create a diagnosability "     +\
                      "context. (This is the theta from the formal definition)."+\
                      " Note: --initial-condition is left unset, the TRUE "     +\
                      "condition is used which means that no constraint is "    +\
                      " imposed on the initial state."                          +\
                      "Warning: No temporal formula is allowed here")


    args.add_argument("-s1", "--sigma1", type=str,
                      help="An LTL formula which is used to constrain the trace"+\
                      " of the first member of the critical pair. Note: this "  +\
                      "option is mutually exclusive with --sigma12")
    args.add_argument("-s2", "--sigma2", type=str,
                      help="An LTL formula which is used to constrain the trace"+\
                      " of the second member of the critical pair. Note: this " +\
                      "option is mutually exclusive with --sigma12")

    args.add_argument("-s12", "--sigma12", type=str,
                      help="The trace part of the diagnosability context. This "+\
                      "option is used to force the diagnosability test to be "  +\
                      "done in a certain context which is considered "          +\
                      "interesting. (Default value: TRUE to not constrain the  "+\
                      "traces in any way). Note: this option is mutually "      +\
                      "exclusive --sigma1 and --sigma2 since it forces both "   +\
                      "sides of the critical pair to satisfy the same trace "   +\
                      "context.")

    parsed_args = args.parse_args()

    # custom vality check
    if (parsed_args.sigma1  is not None) &\
       (parsed_args.sigma2  is not None) &\
       (parsed_args.sigma12 is not None):
        print("Wrong usage: use either (--sigma1 and --sigma2) xor --sigma12")
        exit(1)
    if ((parsed_args.sigma1 is not None) & (parsed_args.sigma2 is None)) \
    or ((parsed_args.sigma1 is None) & (parsed_args.sigma2 is not None)):
        print("Wrong usage: --sigma1 and --sigma2 MUST be used together")
        exit(1)
    if (parsed_args.sigma1  is None) &\
       (parsed_args.sigma2  is None) &\
       (parsed_args.sigma12 is None):
        parsed_args.sigma1 = "TRUE"
        parsed_args.sigma2 = "TRUE"
    else:
        parsed_args.sigma1 = parsed_args.sigma12
        parsed_args.sigma2 = parsed_args.sigma12

    return parsed_args

def mk_observable_names(args):
    """
    returns the set of symbols names that need to be considered observable in
    the context of this diagnosability test.

    .. note::

        The output of this function is typically what need to be passed to
        `mk_observable_vars`

    :param args: the arguments given on the command line
    :return: the list of variable names corresponding to the patterns specified
        by the end user.
    """
    # take
    observable = set(args.observable)

    # handle the --observable-inputs (-oi) command line flag
    if args.observable_inputs:
        scalr_name = lambda x: str(x.scalar)
        input_vars = map(scalr_name, master_be_fsm().encoding.input_variables)
        observable = observable | set(input_vars)

    # handle the external config file (--observable-file // -of) and the
    # multiple --observable-regex (-or) command line options
    regexes = set(args.observable_regex)
    if args.observable_file:
        with open(args.observable_file) as f:
            for line in f:
                regexes = regexes | set(line.split(";"))

    for symbol in master_bool_sexp_fsm().symbols_list:
        for regex in regexes:
            if fullmatch(regex.strip(), str(symbol)):
                observable.add(str(symbol))

    return observable


def mk_observable_vars(var_names):
    """
    Creates the list of the variables (BeVar) that are considered observable in
    the context of this diagnosability test.

    :param var_names: a list of string containing the names of the vars that
        should be visible
    :return: a list of untimed boolean expressions standing for the various bits
        of the observable variables.
    """
    if not var_names:
        return []
    else:
        # otherwise use the user specified set of actions
        encoding         = master_be_fsm().encoding
        boolean_varnames = reduce(lambda x,y: x+y, map(booleanize, var_names), [])
        be_variables     = list(map(lambda x: encoding.by_name[str(x)],boolean_varnames))
        return be_variables

def mk_specs_nodes(spec):
    """
    Creates the Nodes(:see:`pynusmv.node.Node`), that represent the diagnosability
    condition to be verified.

    .. note::

        The diagnosability condition *must* be expressed as a pair of boolean
        expression (propositional formulas) that are separated by a semicolon.

        Example::

            status = active ; status = highlighted

    :param spec: the diagnosability condition to be evaluated.
    :return: a tuple representing the diagnosability condition to be verified.
        (This tuple is composed of two Nodes, which stand for the NuSMV way of
        representing formulas)
    """
    splitted = spec.split(';')
    if len(splitted) < 2:
        raise ValueError("The two parts of the diagnosability condition need to be separated with a semicolon")
    to_node = lambda x: Node.from_ptr(parse_simple_expression(x))
    return tuple(map(to_node, splitted))

def generate_path(offset, length):
    """
    Returns a boolean expression representing a path of length `length` in the
    fsm described by the loaded model.

    :param length: the length of the path in the fsm
    :param offset: the offset at which the path should be starting
    :return: a boolean expression representing a path of length `length` in the
        loaded fsm.
    """
    model = BmcModel()
    path  = model.init[offset] & model.unrolling(offset, offset + length)
    return path

def constraint_same_observations(observable_vars, offset_path1, offset_path2, length):
    """
    Generates a boolean expression stating that the observable state of both
    paths should be the same (all input vars are equivalent).

    :param observable_vars: the list of the boolean variables that are considered
        visible in the scope of this diagnosability test
    :param offset_path1: the offset at which path 1 is supposed to start (should be 0)
    :param offset_path2: the offset at which path 2 is supposed to start (must not intersect with path1)
    :param length: the length of the path
    :return: an expression describing the fact that observations must be the
        exact same along the two paths.
    """
    fsm = master_be_fsm()
    constraint = Be.true(fsm.encoding.manager)
    for time_ in range(length+1):
        for v in observable_vars:
            ep1 = v.at_time[time_ + offset_path1].boolean_expression
            ep2 = v.at_time[time_ + offset_path2].boolean_expression
            constraint &= ep1.iff(ep2)
    return constraint

def constraint_eventually_critical_pair(formula_nodes, offset_path1, offset_path2, length):
    """
    Generates a boolean expression representing the critical pair condition.
    That is to say, it generates a condition that verifies if it is possible that
    the two belief states are inconsistent wrt `formula`.

    :param formula_nodes: the formula whose diagnosability is verified.
    :param offset_path1: the offset at which path 1 is supposed to start (should be 0)
    :param offset_path2: the offset at which path 2 is supposed to start (must not intersect with path1)
    :param length: the length of the path
    :return: an expression describing the 'critical pair' condition.
    """
    enc = master_be_fsm().encoding
    c1  = make_nnf_boolean_wff(formula_nodes[0]).to_be(enc)
    c2  = make_nnf_boolean_wff(formula_nodes[1]).to_be(enc)

    constraint = Be.false(enc.manager)
    for time_ in range(length+1):
        constraint |= ( enc.shift_to_time(c1, time_ + offset_path1)
                      & enc.shift_to_time(c2 , time_ + offset_path2) )
    return constraint

# theta
def constraint_context_theta_initial(initial_condition, offset_path1, offset_path2):
    """
    Generates the theta part of the context constraint. That is to say, the
    part of the constraint which imposes a certain initial condition on the
    initial belief state.

    :param initial_condition: a Node representing the intial condition to be
        enforced on the belief state.
    :param offset_path1: the offset at which path 1 is supposed to start (should be 0)
    :param offset_path2: the offset at which path 2 is supposed to start (must not intersect with path1)
    :return: the theta part of the context constraint encoded in the form of
        a boolean expression.
    """
    enc = master_be_fsm().encoding
    # enforce cond on trace 1
    w01 = enc.shift_to_time(initial_condition.to_be(enc), offset_path1)
    # enforce cond on trace 2
    w02 = enc.shift_to_time(initial_condition.to_be(enc), offset_path2)
    return w01 & w02

def generate_sat_problem(observable_vars, formula_nodes, length, theta, sigma1, sigma2):
    """
    Generates a SAT problem which is satisfiable iff the given `formula` is
    *NOT* diagnosable for the loaded model for traces of length `length`.

    :param observable_vars: the list of the boolean variables that are considered
        visible in the scope of this diagnosability test
    :param formula: the node (NuSMV ast representation) representing the formula
        whose diagnosability is under verification
    :param length: the maximum length of the generated traces.
    :param theta: the initial condition placed on the initial belief state
        (in the form of a :see:`pynusmv.node.Node`)
    :param sigma1: the shape of the traces considered relevant for the first
        member of the critical pair in the ongoing diagnosability test
        (in the form of a :see:`pynusmv.node.Node`)
    :param sigma2: the shape of the traces considered relevant for the second
        member of the critical pair in the ongoing diagnosability test
        (in the form of a :see:`pynusmv.node.Node`)
    :return: a SAT problem which is satisfiable iff the given formula is not
        diagnosable on the loaded model.
    """
    fsm = master_be_fsm()
    offset_1  = 0
    offset_2  = length +1

    problem = generate_path(offset_1, length) & generate_path(offset_2, length) \
            & constraint_same_observations(
                                    observable_vars, offset_1, offset_2, length)\
            & constraint_context_theta_initial(theta, offset_1, offset_2)       \
            & bounded_semantics_at_offset(fsm, sigma1, length, offset_1)        \
            & bounded_semantics_at_offset(fsm, sigma2, length, offset_2)        \
            & constraint_eventually_critical_pair(
                                    formula_nodes, offset_1, offset_2, length)

    return problem

def diagnosability_violation(observable_names, solver, k):
    """
    Interprets the result (model of the sat solver) and prints the parallel
    traces (having the same observations) that lead to some critical pair.

    :param observable_names: the list of names of the variables which are
        considered observable in the model.
    :param solver: the solver that responded SatResult.SATISFIABLE to some
        submitted problem
    :param k: the bound on the length of the problem submitted to the solver.
    """
    lexicographically = lambda x: str(x)
    be_enc = master_be_fsm().encoding
    decoded= be_enc.decode_sat_model(solver.model)

    # create the trace that will actually get returned
    counter_ex = "############### DIAGNOSABILITY VIOLATION ############\n"

    for time in range(k+1):
        counter_ex+= "*************** TIME {:03} ****************************\n".format(time)
        counter_ex+= "--------------- OBSERVABLE STATE --------------------\n"
        # add all observable values.
        for symbol in sorted(decoded[time].keys(), key=lexicographically):
            if str(symbol) in observable_names:
                counter_ex+="{} = {}\n".format(symbol, decoded[time][symbol])

        counter_ex+= "--------------- BELIEF STATE A ----------------------\n"
        # add all non-observable values of the belief STATE A___
        for symbol in sorted(decoded[time].keys(), key=lexicographically):
            if str(symbol) not in observable_names:
                counter_ex+="{} = {}\n".format(symbol, decoded[time][symbol])

        counter_ex+= "--------------- BELIEF STATE B ----------------------\n"
        # add all non-observable values of the belief STATE B___
        for symbol in sorted(decoded[k+1+time].keys(), key=lexicographically):
            if str(symbol) not in observable_names:
                counter_ex+="{} = {}\n".format(symbol, decoded[1+k+time][symbol])

    return counter_ex

def verify_for_size_exactly_k(observable_names, observable_vars, formula_nodes, k, theta, sigma1, sigma2):
    """
    Performs the verification of the diagnosability problem for `formula_node`
    when a maximum of `k` execution steps are allowed.

    :param observable_vars: the list of the boolean variables that are considered
        visible in the scope of this diagnosability test
    :param formula: the node (NuSMV ast representation) representing the formula
        whose diagnosability is under verification
    :param k: the maximum length of the generated traces.
    :param theta: the initial condition placed on the initial belief state
        (in the form of a :see:`pynusmv.node.Node`)
    :param sigma1: the shape of the traces considered relevant for the first
        member of the critical pair in the ongoing diagnosability test
        (in the form of a :see:`pynusmv.node.Node`)
    :param sigma2: the shape of the traces considered relevant for the second
        member of the critical pair in the ongoing diagnosability test
        (in the form of a :see:`pynusmv.node.Node`)
    :return: the text 'No Violation' if no counter example could be found,
        and a counter example when one could be identified.
    """
    problem = generate_sat_problem(observable_vars, formula_nodes, k, theta, sigma1, sigma2)
    problem_= problem.inline(True)  # remove potentially redundant information
    cnf     = problem_.to_cnf()

    solver  = SatSolverFactory.create()
    solver += cnf
    solver.polarity(cnf, Polarity.POSITIVE)

    if solver.solve() == SatSolverResult.SATISFIABLE:
        return diagnosability_violation(observable_names, solver, k)
    else:
        return "No Violation"

def check(args, condition_text, observable):
    """
    Performs the verification of the diagnosability of the condition represented
    by `condition_text` and print its result to stdout.

    :param args: the arguments that were given on the command line
    :param condition_text: a string representing the diagnosability condition to
        be verified in the format 'c1 ; c2'.
    :param observable: the set of symbols considered observable in the context
        of this diagnosability test
    """
    try:
        observable_vars          = mk_observable_vars(observable)
        diagnosability_condition = mk_specs_nodes(condition_text)

        theta = Node.from_ptr(parse_simple_expression(args.initial_condition))
        theta = make_nnf_boolean_wff(theta)

        sigma1= Node.from_ptr(parse_ltl_spec(args.sigma1))
        sigma1= make_nnf_boolean_wff(sigma1).to_node()

        sigma2= Node.from_ptr(parse_ltl_spec(args.sigma2))
        sigma2= make_nnf_boolean_wff(sigma2).to_node()

        for k in range(args.bound+1):
            result = verify_for_size_exactly_k(observable, observable_vars, diagnosability_condition, k, theta, sigma1, sigma2)
            if "No Violation" != str(result):
                print("-- {} is *NOT* diagnosable for length {}".format(diagnosability_condition, k))
                print(result)
                return
            else:
                print("-- No counter example at length {}".format(k))
        print("-- No counter example found for executions of length <= {}".format(args.bound))

    except Exception as e:
        print("The specified condition contains a syntax error")
        print(e)

def print_greeting(model, observable):
    """
    Prints some user greeting information
    """
    print("*"*80)
    print("* MODEL")
    print("*"*80)
    print(model)

    print("*"*80)
    print("* OBSERVABLE VARIABLES")
    print("*"*80)
    print("\n".join(sorted(observable)))


def proceed(args):
    """Actually proceeds to the verification"""
    with init_nusmv():
        load(args.model)
        with BmcSupport():

            observable = mk_observable_names(args)

            if not args.quiet:
                with open(args.model) as m:
                    model = m.read()
                print_greeting(model, observable)

            if args.spec is not None:
                check(args, args.spec, observable)
            else:
                print("*"*80)
                print("* DIAGNOSABILITY TESTS")
                print("*"*80)
                print("Enter diagnosability condition, one per line in the format: 'c1 ; c2'")
                for line in sys.stdin:
                    check(args, line,  observable)

def main():
    proceed(arguments())
    
if __name__ == "__main__":
    main()



from pyparsing import Suppress, SkipTo, Forward, ZeroOrMore, Literal

from .ast import (TrueExp, FalseExp,
                  Atom, Not, And, Or, Implies, Iff, 
                  AaF, AaG, AaX, AaU, AaW,
                  EaF, EaG, EaX, EaU, EaW)
                  
                  
def _left_(clss, tokens):
    """
    Parse tokens and return an AST, assuming left associativity.
    
    Given a list of tokens [v1, op, v2, op, ..., op, vn],
    return res, a hierarchy of instances of clss such that
    res = clss(clss(...clss(v1, v2), ..., vn).
    
    This is a helper function to parse logical operators.
    """
    if len(tokens) == 1:
        return tokens[0]
    else:
        return clss(_left_(clss, tokens[:-2]), tokens[-1])


def _right_(clss, tokens):
    """
    Parse tokens and return an AST, assuming right associativity.
    
    Given a list of tokens [v1, op, v2, op, ..., op, vn],
    return res, a hierarchy of instances of clss such that
    res = clss(v1, clss(v2, ..., vn)...).
    
    This is a helper function to parse logical operators.
    """
    if len(tokens) == 1:
        return tokens[0]
    else:
        return clss(tokens[0], _right_(clss, tokens[2:]))
        
        
def _logicals_(atomic):
    """
    Return a new parser parsing logical expressions on atomics.
    
    This parser recognizes the following grammar, with precedences
    parser := atomic | '~' parser | parser '&' parser | parser '|' parser |
              parser '->' parser | parser '<->' parser
              
    Returned AST uses .ast package's classes.
    """
    parser = Forward()

    atom = (atomic | Suppress("(") + parser + Suppress(")"))

    notstrict = "~" + atom
    notstrict.setParseAction(lambda tokens: Not(tokens[1]))
    not_ = notstrict | atom
    and_ = not_ + ZeroOrMore("&" + not_)
    and_.setParseAction(lambda tokens: _left_(And, tokens))
    or_ = and_ + ZeroOrMore("|" + and_)
    or_.setParseAction(lambda tokens: _left_(Or, tokens))
    implies = ZeroOrMore(or_ + "->") + or_
    implies.setParseAction(lambda tokens: _right_(Implies, tokens))
    iff = implies + ZeroOrMore("<->" + implies)
    iff.setParseAction(lambda tokens: _left_(Iff, tokens))

    parser <<= iff
    
    return parser
        

"""
ARCTL parsing tool.

_arctl       := _atom | _logical | _temporal
_logical     := '~' __arctl | '(' _logical ')' | __arctl '&' __arctl |
                __arctl '|' __arctl | __arctl '->' __arctl | __arctl '<->' __arctl
_temporal    := 'A' '<' _action '>' 'F' __arctl |
                'A' '<' _action '>' 'G' __arctl |
                'A' '<' _action '>' 'X' __arctl |
                'A' '<' _action '>' '[' __arctl 'U' __arctl ']' |
                'A' '<' _action '>' '[' __arctl 'W' __arctl ']' |
                'E' '<' _action '>' 'F' __arctl |
                'E' '<' _action '>' 'G' __arctl |
                'E' '<' _action '>' 'X' __arctl |
                'E' '<' _action '>' '[' __arctl 'U' __arctl ']' |
                'E' '<' _action '>' '[' __arctl 'W' __arctl ']'
_action      := _atom | '(' _action ')' | '~' _action | _action '&' _action |
                _action '|' _action | _action '->' _action |
                _action '<->' _action
               
_atom is defined by any string surrounded by single quotes.

_action and _logical are specified with usual precedences and associativity,
i.e.
prec : ~, &, |, ->, <->
assoc : &, |, <-> left assoc, ->, ~ right assoc


The parser returns a structure embedding the structure of the parsed
expression, represented using AST classes of .ast module.
"""

__arctl = None

def parseArctl(spec):
    """Parse the spec and return its AST."""
    global __arctl
    if __arctl is None:
        true = Literal("True")
        true.setParseAction(lambda tokens: TrueExp())
        false = Literal("False")
        false.setParseAction(lambda tokens: FalseExp())
        atom = "'" + SkipTo("'") + "'"
        atom.setParseAction(lambda tokens: Atom(tokens[1]))
        
        action = _logicals_(atom)
        
        __arctl = Forward()
        
        proposition = true | false | atom

        notproposition = "~" + proposition
        notproposition.setParseAction(lambda tokens: Not(tokens[1]))
        formula = (proposition | notproposition |
                   Suppress("(") + __arctl + Suppress(")"))


        temporal = Forward()

        e = Literal("E") + "<" + action + ">"
        a = Literal("A") + "<" + action + ">"

        eax = e + "X" + temporal
        eax.setParseAction(lambda tokens: EaX(tokens[2], tokens[5]))
        aax = a + "X" + temporal
        aax.setParseAction(lambda tokens: AaX(tokens[2], tokens[5]))
        
        eaf = e + "F" + temporal
        eaf.setParseAction(lambda tokens: EaF(tokens[2], tokens[5]))
        aaf = a + "F" + temporal
        aaf.setParseAction(lambda tokens: AaF(tokens[2], tokens[5]))
        
        eag = e + "G" + temporal
        eag.setParseAction(lambda tokens: EaG(tokens[2], tokens[5]))
        aag = a + "G" + temporal
        aag.setParseAction(lambda tokens: AaG(tokens[2], tokens[5]))
        
        eau = e + "[" + __arctl + "U" + __arctl + "]"
        eau.setParseAction(lambda tokens: EaU(tokens[2], tokens[5], tokens[7]))
        aau = a + "[" + __arctl + "U" + __arctl + "]"
        aau.setParseAction(lambda tokens: AaU(tokens[2], tokens[5], tokens[7]))
        
        eaw = e + "[" + __arctl + "W" + __arctl + "]"
        eaw.setParseAction(lambda tokens: EaW(tokens[2], tokens[5], tokens[7]))
        aaw = a + "[" + __arctl + "W" + __arctl + "]"
        aaw.setParseAction(lambda tokens: AaW(tokens[2], tokens[5], tokens[7]))


        temporal <<= (formula | eax | aax | eaf | aaf | eag | aag |
                      eau | aau | eaw | aaw)
      
        logical = _logicals_(temporal)

        __arctl <<= logical
    
    return __arctl.parseString(spec, parseAll = True)


from pynusmv_lower_interface.nusmv.parser import parser
from pynusmv.dd import BDD
from pynusmv.mc import eval_ctl_spec
from pynusmv.utils import fixpoint

from pynusmv_lower_interface.nusmv.fsm.bdd import bdd as nsBddFsm

def check(fsm, spec, context=None):
    """
    Return whether spec in context is satisfied by fsm.
    """
    violating = fsm.init & ~eval_ctl(fsm, spec, context=context)
    return violating.is_false()

def eval_ctl(fsm, spec, context = None):
    """
    Evaluate spec in fsm.
    
    Return the BDD representing all states of fsm satisfying spec.
    """
    
    if spec.type == parser.CONTEXT:
        return eval_ctl(fsm, spec.cdr, spec.car)
        
    elif spec.type == parser.FALSEEXP:
        return BDD.false(fsm.bddEnc.DDmanager)
        
    elif spec.type == parser.TRUEEXP:
        return BDD.true(fsm.bddEnc.DDmanager)
        
    elif spec.type == parser.NOT:
        return ~eval_ctl(fsm, spec.car, context)
        
    elif spec.type == parser.OR:
        left = eval_ctl(fsm, spec.car, context)
        right = eval_ctl(fsm, spec.cdr, context)
        return left | right
    
    elif spec.type == parser.AND:
        left = eval_ctl(fsm, spec.car, context)
        right = eval_ctl(fsm, spec.cdr, context)
        return left & right
    
    elif spec.type == parser.IMPLIES:
        left = eval_ctl(fsm, spec.car, context)
        right = eval_ctl(fsm, spec.cdr, context)
        return ~left | right
                            
    elif spec.type == parser.IFF:
        left = eval_ctl(fsm, spec.car, context)
        right = eval_ctl(fsm, spec.cdr, context)
        return (left & right) | (~left & ~right)
                    
    elif spec.type == parser.EX:
        return ex(fsm, eval_ctl(fsm, spec.car, context))
                                 
    elif spec.type == parser.EF:
        left = BDD.true(fsm.bddEnc.DDmanager)
        right = eval_ctl(fsm, spec.car, context)
        return eu(fsm, left, right)
                                 
    elif spec.type == parser.EG:
        return eg(fsm, eval_ctl(fsm, spec.car, context))
                                 
    elif spec.type == parser.EU:
        return eu(fsm,
                  eval_ctl(fsm, spec.car, context),
                  eval_ctl(fsm, spec.cdr, context))
                    
    elif spec.type == parser.EW:
        left = eval_ctl(fsm, spec.car, context)
        right = eval_ctl(fsm, spec.cdr, context)
        return eg(fsm, left) | eu(fsm, left, right)
                    
    elif spec.type == parser.AX:
        left = eval_ctl(fsm, spec.car, context)
        return ~ex(fsm, ~left)        
        
    elif spec.type == parser.AF:
        left = eval_ctl(fsm, spec.car, context)
        return ~eg(fsm, ~left)
                                 
    elif spec.type == parser.AG:
        left = eval_ctl(fsm, spec.car, context)
        true = BDD.true(fsm.bddEnc.DDmanager)
        return ~eu(fsm, true, ~left)
                                 
    elif spec.type == parser.AU:
        # A[p U q] = ¬E[¬q W (¬p & ¬q)] = ¬(E[¬q U (¬p & ¬q)] | EG ¬q)
        left = eval_ctl(fsm, spec.car, context)
        right = eval_ctl(fsm, spec.cdr, context)
        return ~(eu(fsm, ~right, ~left & ~right) | eg(fsm, ~right))
                        
    elif spec.type == parser.AW:
        left = eval_ctl(fsm, spec.car, context)
        right = eval_ctl(fsm, spec.cdr, context)
        return ~eu(fsm, ~right, ~left & ~right)
    
    else:
        return eval_ctl_spec(fsm, spec)
        
        
def ex(fsm, phi):
    phi = phi & fair_states(fsm) & fsm.reachable_states
    return fsm.pre(phi) & fsm.reachable_states
    
    
def eg(fsm, phi):    
    # EG p = nu Z . p & &_(f in F) Pre( mu Y . (Z & f) | (p & Pre(Y)) )
    #      = nu Z . p & &_(f in F) EX( mu Y . (Z & f) | (p & EX(Y)))
    
    def inner(Z):
        res = Z
        for f in fsm.fairness_constraints:
            res = res & fixpoint(lambda Y : (Z & f) | (phi & 
                                fsm.weak_pre(Y.forsome(fsm.bddEnc.inputsCube))),
                                 BDD.false(fsm.bddEnc.DDmanager))
        return phi & fsm.weak_pre(res.forsome(fsm.bddEnc.inputsCube))
        
    r = fixpoint(inner, BDD.true(fsm.bddEnc.DDmanager))
    return r.forsome(fsm.bddEnc.inputsCube)
    
    
def eu(fsm, phi, psi):
    # E[p U q] = q | (p & EX E[p U q]) = mu Z . q | (p & Pre(Z))
    return fixpoint(lambda X : (psi & fair_states(fsm) & fsm.reachable_states) |
                               (phi & ex(fsm, X)),
                    BDD.false(fsm.bddEnc.DDmanager))
    
    
def fair_states(fsm):
    return eg(fsm, BDD.true(fsm.bddEnc.DDmanager))


import unittest

from pynusmv.init import init_nusmv, deinit_nusmv
from pynusmv.mc import eval_simple_expression

from pynusmv_tools.mas import glob
from pynusmv_tools.ctlk.eval import evalCTLK, ex, eu, eg, nk, nd, ne, nc
from pynusmv_tools.ctlk.parsing import parseCTLK

from pynusmv_tools.ctlk.explain import (explain_ex, explain_eg, explain_eu,
                                explain_nk, explain_ne, explain_nd, explain_nc,
                                explain_reachable)

class TestExplain(unittest.TestCase):
    
    def setUp(self):
        init_nusmv()
        
    def tearDown(self):
        glob.reset_globals()
        deinit_nusmv()
        
    def model(self):
        glob.load_from_file("tests/pynusmv_tools/ctlk/dining-crypto.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
        
    def simplemodel(self):
        glob.load_from_file("tests/pynusmv_tools/ctlk/agents.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
        
    def test_ex_simple(self):
        fsm = self.simplemodel()
        
        lt = eval_simple_expression(fsm, "at.local")
        lf = eval_simple_expression(fsm, "af.local")
        g = eval_simple_expression(fsm, "global")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        # lt & lf & g |= EX !lt
        # because (lt & lf & g, !lt & lf (g | !g))
        state = fsm.pick_one_state(lt & lf & g)
        witness = explain_ex(fsm, state, ~lt)
        self.assertTrue(witness[0] == state)
        self.assertTrue(witness[2] <= ~lt)
        self.assertTrue(witness[2] <= ~lt & lf)
        self.assertTrue((witness[2] == ~lt & lf & g) |
                        (witness[2] == ~lt & lf & ~g))
        self.assertTrue(fsm.get_inputs_between_states(state, witness[2])
                                                                 .isnot_false())
        self.assertTrue(witness[1] <= fsm.get_inputs_between_states(state,
                                                                    witness[2]))
                                                                    
                                                                    
    def test_ex_dincry(self):
        fsm = self.model()
        
        c1p = eval_simple_expression(fsm, "c1.payer")
        c2p = eval_simple_expression(fsm, "c2.payer")
        c3p = eval_simple_expression(fsm, "c3.payer")
        c1h = eval_simple_expression(fsm, "c1.coin = head")
        c2h = eval_simple_expression(fsm, "c2.coin = head")
        c3h = eval_simple_expression(fsm, "c3.coin = head")
        odd = eval_simple_expression(fsm, "countsay = odd")
        even = eval_simple_expression(fsm, "countsay = even")
        unk = eval_simple_expression(fsm, "countsay = unknown")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        # unknown & c1h & c2h & c3h |= odd
        # because (unk & c1h & c2h & c3h & c1p & !c2p & !c3p,
        #          odd & c1h & c2h & c3h & c1p & !c2p & !c3p)
        state = fsm.pick_one_state(unk & c1h & c2h & c3h & c1p)
        self.assertTrue(state <= ex(fsm, odd))
        self.assertTrue(state <= (unk & c1h & c2h & c3h & c1p))
        self.assertTrue(state <= fsm.init)
        witness = explain_ex(fsm, state, odd)
        self.assertTrue(witness[0] == state)
        self.assertTrue(witness[1].isnot_false())
        self.assertTrue(witness[2] <= odd)
        self.assertTrue(witness[2] <= odd & c1p)
        self.assertTrue(fsm.get_inputs_between_states(state, witness[2])
                                                                 .isnot_false())
        self.assertTrue(witness[1] <= fsm.get_inputs_between_states(state,
                                                                    witness[2]))
                                                                    
                                                                    
    def test_eu_simple(self):
        fsm = self.simplemodel()
        
        lt = eval_simple_expression(fsm, "at.local")
        lf = eval_simple_expression(fsm, "af.local")
        g = eval_simple_expression(fsm, "global")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        # lt & lf & g |= E[ g U ~lf & ~lt & ~g ]
        eus = eu(fsm, g, ~lf & ~lt & ~g)
        state = fsm.pick_one_state(eus)
        witness = explain_eu(fsm, state, g, ~lf & ~lt & ~g)
        self.assertTrue(witness[0] == state)
        self.assertTrue(witness[0] <= g | ~lf & ~lt & ~g)
        
        for (s, i, sp) in zip(witness[:-2:2], witness[1:-2:2], witness[2:-2:2]):
            self.assertTrue(s <= g)
            self.assertTrue(sp <= g)
            self.assertIsNotNone(i)
            self.assertTrue(i <= fsm.get_inputs_between_states(s, sp))        
        
        self.assertIsNotNone(witness[-2])
        self.assertTrue(witness[-2] <=
                        fsm.get_inputs_between_states(witness[-3], witness[-1]))
        self.assertTrue(witness[-1] <= ~lf)
        self.assertTrue(witness[-1] <= ~lt)
        self.assertTrue(witness[-1] <= ~g)
        self.assertTrue(witness[-1] == ~lf & ~lt & ~g)
                                                                    
                                                                    
    def test_eu_dincry(self):
        fsm = self.model()
        
        c1p = eval_simple_expression(fsm, "c1.payer")
        c2p = eval_simple_expression(fsm, "c2.payer")
        c3p = eval_simple_expression(fsm, "c3.payer")
        c1h = eval_simple_expression(fsm, "c1.coin = head")
        c2h = eval_simple_expression(fsm, "c2.coin = head")
        c3h = eval_simple_expression(fsm, "c3.coin = head")
        odd = eval_simple_expression(fsm, "countsay = odd")
        even = eval_simple_expression(fsm, "countsay = even")
        unk = eval_simple_expression(fsm, "countsay = unknown")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        # init |= E[c1p U odd]
        eus = eu(fsm, c1p, odd)
        self.assertTrue((eus & fsm.init).isnot_false())
        state = fsm.pick_one_state(eus & fsm.init)
        witness = explain_eu(fsm, state, c1p, odd)
        self.assertTrue(witness[0] == state)
        self.assertTrue(witness[0] <= c1p | odd)
        
        for (s, i, sp) in zip(witness[:-2:2], witness[1:-2:2], witness[2:-2:2]):
            self.assertTrue(s <= c1p)
            self.assertTrue(sp <= c1p)
            self.assertIsNotNone(i)
            self.assertTrue(i <= fsm.get_inputs_between_states(s, sp))        
        
        self.assertIsNotNone(witness[-2])
        self.assertTrue(witness[-2] <=
                        fsm.get_inputs_between_states(witness[-3], witness[-1]))
        self.assertTrue(witness[-1] <= odd)
        
    def test_eg_simple(self):
        fsm = self.simplemodel()
        
        lt = eval_simple_expression(fsm, "at.local")
        lf = eval_simple_expression(fsm, "af.local")
        g = eval_simple_expression(fsm, "global")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        # lt & lf & !g |= EG !g
        egs = eg(fsm, ~g)
        self.assertTrue(egs.isnot_false())
        state = fsm.pick_one_state(egs)
        (path, loop) = explain_eg(fsm, state, ~g)
        self.assertTrue(path[0] == state)
        self.assertTrue(path[0] <= ~g & egs)
        
        for (s, i, sp) in zip(path[::2], path[1::2], path[2::2]):
            self.assertTrue(s <= ~g)
            self.assertTrue(sp <= ~g)
            self.assertIsNotNone(i)
            self.assertTrue(i <= fsm.get_inputs_between_states(s, sp))
            
        self.assertIsNotNone(loop)
        self.assertEqual(len(loop), 2)
        self.assertIsNotNone(loop[0])
        self.assertIsNotNone(loop[1])
        
        self.assertTrue(loop[1] in path)
        self.assertTrue(loop[0] <=
                               fsm.get_inputs_between_states(path[-1], loop[1]))
                                                                    
                                                                    
    def test_eg_dincry(self):
        fsm = self.model()
        
        c1p = eval_simple_expression(fsm, "c1.payer")
        c2p = eval_simple_expression(fsm, "c2.payer")
        c3p = eval_simple_expression(fsm, "c3.payer")
        c1h = eval_simple_expression(fsm, "c1.coin = head")
        c2h = eval_simple_expression(fsm, "c2.coin = head")
        c3h = eval_simple_expression(fsm, "c3.coin = head")
        odd = eval_simple_expression(fsm, "countsay = odd")
        even = eval_simple_expression(fsm, "countsay = even")
        unk = eval_simple_expression(fsm, "countsay = unknown")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        # init |= EG c2.payer
        egs = eg(fsm, c2p)
        self.assertTrue((egs & fsm.init).isnot_false())
        state = fsm.pick_one_state(egs & fsm.init)
        (path, loop) = explain_eg(fsm, state, c2p)
        self.assertTrue(path[0] == state)
        self.assertTrue(path[0] <= c2p)
        
        for (s, i, sp) in zip(path[::2], path[1::2], path[2::2]):
            self.assertTrue(s <= c2p)
            self.assertTrue(sp <= c2p)
            self.assertIsNotNone(i)
            self.assertTrue(i <= fsm.get_inputs_between_states(s, sp))
            
        self.assertIsNotNone(loop)
        self.assertEqual(len(loop), 2)
        self.assertIsNotNone(loop[0])
        self.assertIsNotNone(loop[1])
        
        self.assertTrue(loop[1] in path)
        self.assertTrue(loop[0] <=
                               fsm.get_inputs_between_states(path[-1], loop[1]))
                               
                               
    def test_nk_simple(self):
        fsm = self.simplemodel()
        
        lt = eval_simple_expression(fsm, "at.local")
        lf = eval_simple_expression(fsm, "af.local")
        g = eval_simple_expression(fsm, "global")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        # nK<at> af.local
        nks = nk(fsm, "at", lf)
        self.assertTrue(nks.isnot_false())
        state = fsm.pick_one_state(nks)
        (s, ag, sp) = explain_nk(fsm, state, "at", lf)
        self.assertEqual(s, state)
        self.assertEqual(ag, {"at"})
        self.assertTrue(sp <= lf)
        self.assertTrue(sp <= fsm.equivalent_states(state, {"at"}))
        self.assertTrue(sp <= fsm.reachable_states)
                                   
                                                                    
    def test_nk_dincry(self):
        fsm = self.model()
        
        c1p = eval_simple_expression(fsm, "c1.payer")
        c2p = eval_simple_expression(fsm, "c2.payer")
        c3p = eval_simple_expression(fsm, "c3.payer")
        c1h = eval_simple_expression(fsm, "c1.coin = head")
        c2h = eval_simple_expression(fsm, "c2.coin = head")
        c3h = eval_simple_expression(fsm, "c3.coin = head")
        odd = eval_simple_expression(fsm, "countsay = odd")
        even = eval_simple_expression(fsm, "countsay = even")
        unk = eval_simple_expression(fsm, "countsay = unknown")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        # nK<c2> ~c3.payer
        nks = nk(fsm, "c2", ~c3p)
        self.assertTrue(nks.isnot_false())
        state = fsm.pick_one_state(nks)
        (s, ag, sp) = explain_nk(fsm, state, "c2", ~c3p)
        self.assertEqual(s, state)
        self.assertEqual(ag, {"c2"})
        self.assertTrue(sp <= ~c3p)
        self.assertTrue(sp <= fsm.equivalent_states(state, {"c2"}))
        self.assertTrue(sp <= fsm.reachable_states)
        
        
    def test_ne_simple(self):
        fsm = self.simplemodel()
        
        lt = eval_simple_expression(fsm, "at.local")
        lf = eval_simple_expression(fsm, "af.local")
        g = eval_simple_expression(fsm, "global")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        # nE<at, af> g
        nes = ne(fsm, ["at", "af"], g)
        self.assertTrue(nes.isnot_false())
        state = fsm.pick_one_state(nes)
        (s, ag, sp) = explain_ne(fsm, state, ["at", "af"], g)
        self.assertEqual(s, state)
        self.assertTrue(ag in [{"at"}, {"af"}])
        self.assertTrue(sp <= g)
        self.assertTrue((sp <= fsm.equivalent_states(state, {"at"})) |
                        (sp <= fsm.equivalent_states(state, {"af"})))
        self.assertTrue(sp <= fsm.reachable_states)
                                   
                                                                    
    def test_ne_dincry(self):
        fsm = self.model()
        
        c1p = eval_simple_expression(fsm, "c1.payer")
        c2p = eval_simple_expression(fsm, "c2.payer")
        c3p = eval_simple_expression(fsm, "c3.payer")
        c1h = eval_simple_expression(fsm, "c1.coin = head")
        c2h = eval_simple_expression(fsm, "c2.coin = head")
        c3h = eval_simple_expression(fsm, "c3.coin = head")
        odd = eval_simple_expression(fsm, "countsay = odd")
        even = eval_simple_expression(fsm, "countsay = even")
        unk = eval_simple_expression(fsm, "countsay = unknown")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        # nE<c1,c2> ~c3.payer
        nes = ne(fsm, ["c1", "c2"], ~c3p)
        self.assertTrue(nes.isnot_false())
        state = fsm.pick_one_state(nes)
        (s, ag, sp) = explain_ne(fsm, state, ["c1", "c2"], ~c3p)
        self.assertEqual(s, state)
        self.assertTrue(ag in [{"c1"}, {"c2"}])
        self.assertTrue(sp <= ~c3p)
        self.assertTrue((sp <= fsm.equivalent_states(state, {"c1"})) |
                        (sp <= fsm.equivalent_states(state, {"c2"})))
        self.assertTrue(sp <= fsm.reachable_states)
        

    def test_nd_simple(self):
        fsm = self.simplemodel()
        
        lt = eval_simple_expression(fsm, "at.local")
        lf = eval_simple_expression(fsm, "af.local")
        g = eval_simple_expression(fsm, "global")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        # nD<at, af> at.local
        nds = nd(fsm, ["at", "af"], lt)
        self.assertTrue(nds.isnot_false())
        state = fsm.pick_one_state(nds)
        (s, ag, sp) = explain_nd(fsm, state, ["at", "af"], lt)
        self.assertEqual(s, state)
        self.assertEqual(ag, {"at", "af"})
        self.assertTrue(sp <= lt)
        self.assertTrue((sp <= fsm.equivalent_states(state, {"at"})) &
                        (sp <= fsm.equivalent_states(state, {"af"})))
        self.assertTrue(sp <= fsm.reachable_states)
                                   
                                                                    
    def test_nd_dincry(self):
        fsm = self.model()
        
        c1p = eval_simple_expression(fsm, "c1.payer")
        c2p = eval_simple_expression(fsm, "c2.payer")
        c3p = eval_simple_expression(fsm, "c3.payer")
        c1h = eval_simple_expression(fsm, "c1.coin = head")
        c2h = eval_simple_expression(fsm, "c2.coin = head")
        c3h = eval_simple_expression(fsm, "c3.coin = head")
        odd = eval_simple_expression(fsm, "countsay = odd")
        even = eval_simple_expression(fsm, "countsay = even")
        unk = eval_simple_expression(fsm, "countsay = unknown")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        # nD<c1,c2> ~c3.payer
        nds = nd(fsm, ["c1", "c2"], ~c3p)
        self.assertTrue(nds.isnot_false())
        state = fsm.pick_one_state(nds)
        (s, ag, sp) = explain_nd(fsm, state, ["c1", "c2"], ~c3p)
        self.assertEqual(s, state)
        self.assertEqual(ag, {"c1", "c2"})
        self.assertTrue(sp <= ~c3p)
        self.assertTrue((sp <= fsm.equivalent_states(state, {"c1"})) &
                        (sp <= fsm.equivalent_states(state, {"c2"})))
        self.assertTrue(sp <= fsm.reachable_states)
  
      
    def test_nc_simple(self):
        fsm = self.simplemodel()
        
        lt = eval_simple_expression(fsm, "at.local")
        lf = eval_simple_expression(fsm, "af.local")
        g = eval_simple_expression(fsm, "global")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        # ~lt & ~lf & g |= nc<at, af> ~lt & ~lf & ~g
        ncs = nc(fsm, ["at"], ~lt & ~lf & g)
        self.assertTrue(ncs.isnot_false())
        state = fsm.pick_one_state(~lt & lf & g)
        self.assertTrue(state <= ncs)
        witness = explain_nc(fsm, state, ["at"], ~lt & ~lf & g)
        self.assertTrue(witness[0] == state)
        for (s, ag, sp) in zip(witness[::2], witness[1::2], witness[2::2]):
            self.assertTrue(s.isnot_false())
            self.assertTrue(ag in [{"at"}])           
            self.assertTrue(sp.isnot_false())
            self.assertTrue(sp <= fsm.equivalent_states(s, {"at"}))
            
        self.assertTrue(witness[-1] <= ~lt & ~lf & g)
                                   
                                                                    
    def test_nc_dincry(self):
        fsm = self.model()
        
        c1p = eval_simple_expression(fsm, "c1.payer")
        c2p = eval_simple_expression(fsm, "c2.payer")
        c3p = eval_simple_expression(fsm, "c3.payer")
        c1h = eval_simple_expression(fsm, "c1.coin = head")
        c2h = eval_simple_expression(fsm, "c2.coin = head")
        c3h = eval_simple_expression(fsm, "c3.coin = head")
        odd = eval_simple_expression(fsm, "countsay = odd")
        even = eval_simple_expression(fsm, "countsay = even")
        unk = eval_simple_expression(fsm, "countsay = unknown")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        # nC<c1,c2> ~c3.payer
        ncs = nc(fsm, ["c1", "c2"], c3h)
        self.assertTrue(ncs.isnot_false())
        state = fsm.pick_one_state(ncs)
        witness = explain_nc(fsm, state, ["c1", "c2"], c3h)
        self.assertTrue(witness[0] == state)
        for (s, ag, sp) in zip(witness[::2], witness[1::2], witness[2::2]):
            self.assertTrue(s.isnot_false())
            self.assertTrue(ag in [{"c1"}, {"c2"}])
            self.assertTrue(sp.isnot_false())
            self.assertTrue((sp <= fsm.equivalent_states(s, {"c1"})) |
                            (sp <= fsm.equivalent_states(s, {"c2"})))
            
        self.assertTrue(witness[-1] <= c3h)
        
        
    def test_reachable_simple(self):
        fsm = self.simplemodel()
        
        lt = eval_simple_expression(fsm, "at.local")
        lf = eval_simple_expression(fsm, "af.local")
        g = eval_simple_expression(fsm, "global")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        state = fsm.pick_one_state(~lf & ~lt & ~g)
        self.assertTrue(state <= fsm.reachable_states)
        witness = explain_reachable(fsm, state)
        self.assertIsNotNone(witness)
        self.assertEqual(witness[0], state)
        
        for (s, i, sp) in zip(witness[::2], witness[1::2], witness[2::2]):
            self.assertTrue(s <= fsm.reachable_states)
            self.assertTrue(sp <= fsm.reachable_states)
            self.assertIsNotNone(i)
            self.assertTrue(i <= fsm.get_inputs_between_states(sp, s))
        
        self.assertTrue(witness[-1] <= fsm.init)
        
        
    def test_reachable_dincry(self):
        fsm = self.model()
        
        c1p = eval_simple_expression(fsm, "c1.payer")
        c2p = eval_simple_expression(fsm, "c2.payer")
        c3p = eval_simple_expression(fsm, "c3.payer")
        c1h = eval_simple_expression(fsm, "c1.coin = head")
        c2h = eval_simple_expression(fsm, "c2.coin = head")
        c3h = eval_simple_expression(fsm, "c3.coin = head")
        odd = eval_simple_expression(fsm, "countsay = odd")
        even = eval_simple_expression(fsm, "countsay = even")
        unk = eval_simple_expression(fsm, "countsay = unknown")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        state = fsm.pick_one_state(c1p & ~c2p & ~c3p & odd)
        self.assertTrue(state <= fsm.reachable_states)
        witness = explain_reachable(fsm, state)
        self.assertIsNotNone(witness)
        self.assertEqual(witness[0], state)
        
        for (s, i, sp) in zip(witness[::2], witness[1::2], witness[2::2]):
            self.assertTrue(s <= fsm.reachable_states)
            self.assertTrue(sp <= fsm.reachable_states)
            self.assertIsNotNone(i)
            self.assertTrue(i <= fsm.get_inputs_between_states(sp, s))
        
        self.assertTrue(witness[-1] <= fsm.init)


"""
AST classes for CTLK formulas representation.
"""

from collections import namedtuple

class Spec:
    """A Spec is represents a CTLK formula."""

# TrueExp, FalseExp, Init, Reachable
TrueExp = namedtuple('TrueExp', [])
TrueExp.__bases__ += (Spec,)
TrueExp.__str__ = lambda self: "True"

FalseExp = namedtuple('FalseExp', [])
FalseExp.__bases__ += (Spec,)
FalseExp.__str__ = lambda self: "False"

Init = namedtuple('Init', [])
Init.__bases__ += (Spec,)
Init.__str__ = lambda self: "Init"

Reachable = namedtuple('Reachable', [])
Reachable.__bases__ += (Spec,)
Reachable.__str__ = lambda self: "Reachable"

# Atom
Atom = namedtuple('Atom', ['value'])
Atom.__bases__ += (Spec,)
Atom.__str__ = lambda self: "'" + str(self.value) + "'"

# logical : Not, And, Or, Implies, Iff
Not = namedtuple('Not', ['child'])
Not.__bases__ += (Spec,)
Not.__str__ = lambda self: "~" + "(" + str(self.child) + ")"

And = namedtuple('And', ['left', 'right'])
And.__bases__ += (Spec,)
And.__str__ = lambda self: "(" + str(self.left) + " & " + str(self.right) + ")"

Or = namedtuple('Or', ['left', 'right'])
Or.__bases__ += (Spec,)
Or.__str__ = lambda self: "(" + str(self.left) + " | " + str(self.right) + ")"

Implies = namedtuple('Implies', ['left', 'right'])
Implies.__bases__ += (Spec,)
Implies.__str__ = (lambda self: "(" + str(self.left) + " -> " +
                                      str(self.right) + ")")

Iff = namedtuple('Iff', ['left', 'right'])
Iff.__bases__ += (Spec,)
Iff.__str__ = (lambda self: "(" + str(self.left) + " <-> " +
                                  str(self.right) + ")")
                                  

# temporal : AF, AG, AX, AU, AW, EF, EG, EX, EU, EW
AF = namedtuple('AF', ['child'])
AF.__bases__ += (Spec,)
AF.__str__ = (lambda self: "A" + "F " + str(self.child))
                            
AG = namedtuple('AG', ['child'])
AG.__bases__ += (Spec,)
AG.__str__ = (lambda self: "A" + "G " + str(self.child))
                           
AX = namedtuple('AX', ['child'])
AX.__bases__ += (Spec,)
AX.__str__ = (lambda self: "A" + "X " + str(self.child))
                           
AU = namedtuple('AU', ['left', 'right'])
AU.__bases__ += (Spec,)
AU.__str__ = (lambda self: "A" + "[" + str(self.left) +
                                 " U " + str(self.right) + "]")
                           
AW = namedtuple('EW', ['left', 'right'])
AW.__bases__ += (Spec,)
AW.__str__ = (lambda self: "A"+ "[" + str(self.left) +
                                " W " + str(self.right) + "]")
 
EF = namedtuple('EF', ['child'])
EF.__bases__ += (Spec,)
EF.__str__ = (lambda self: "E" + "F " + str(self.child))
                           
EG = namedtuple('EG', ['child'])
EG.__bases__ += (Spec,)
EG.__str__ = (lambda self: "E" + "G " + str(self.child))
                           
EX = namedtuple('EX', ['child'])
EX.__bases__ += (Spec,)
EX.__str__ = (lambda self: "E" + "X " + str(self.child))
                           
EU = namedtuple('EU', ['left', 'right'])
EU.__bases__ += (Spec,)
EU.__str__ = (lambda self: "E"+ "[" + str(self.left) +
                                " U " + str(self.right) + "]")
                           
EW = namedtuple('EW', ['left', 'right'])
EW.__bases__ += (Spec,)
EW.__str__ = (lambda self: "E"+ "[" + str(self.left) +
                                " W " + str(self.right) + "]")
                                

# epistemic : nK, nE, nD, nC, K, E, D, C
nK = namedtuple('nK', ['agent', 'child'])
nK.__bases__ += (Spec,)
nK.__str__ = (lambda self: "nK" + "<" + str(self.agent) + ">" + " "
                                + str(self.child))

nE = namedtuple('nE', ['group', 'child'])
nE.__bases__ += (Spec,)
nE.__str__ = (lambda self: "nE" + "<" +
                           ','.join([ag.value for ag in self.group]) +
                           ">" + " " + str(self.child))

nD = namedtuple('nD', ['group', 'child'])
nD.__bases__ += (Spec,)
nD.__str__ = (lambda self: "nD" + "<" +
                           ','.join([ag.value for ag in self.group]) +
                           ">" + " " + str(self.child))

nC = namedtuple('nC', ['group', 'child'])
nC.__bases__ += (Spec,)
nC.__str__ = (lambda self: "nC" + "<" +
                           ','.join([ag.value for ag in self.group]) +
                           ">" + " " + str(self.child))
                                
K = namedtuple('K', ['agent', 'child'])
K.__bases__ += (Spec,)
K.__str__ = (lambda self: "K" + "<" + str(self.agent) + ">" + " "
                              + str(self.child))

E = namedtuple('E', ['group', 'child'])
E.__bases__ += (Spec,)
E.__str__ = (lambda self: "E" + "<" +
                           ','.join([ag.value for ag in self.group]) +
                           ">" + " " + str(self.child))

D = namedtuple('D', ['group', 'child'])
D.__bases__ += (Spec,)
D.__str__ = (lambda self: "D" + "<" +
                           ','.join([ag.value for ag in self.group]) +
                           ">" + " " + str(self.child))

C = namedtuple('C', ['group', 'child'])
C.__bases__ += (Spec,)
C.__str__ = (lambda self: "C" + "<" +
                           ','.join([ag.value for ag in self.group]) +
                           ">" + " " + str(self.child))


"""
This package provides a checker for NuSMV models and CTL properties without
fairness. This means that, even the model provides some fairness constraints,
CTL is checked without taking them into account.
"""


"""ARCTL TLACE explanations."""

from ..eval import evalArctl
from ..ast import (TrueExp, FalseExp, Atom, Not, And, Or, Implies, Iff,
                   AaF, AaG, AaX, AaU, AaW, EaF, EaG, EaX, EaU, EaW)
from .tlace import (Tlacenode, Tlacebranch)
from ..explain import explain_eax, explain_eag, explain_eau

def explain_witness(fsm, state, spec):
    """
    Explain why state of fsm satisfies spec.
    
    Return a TLACE node explaining why state of fsm satifies spec.
    """
    
    if type(spec) is TrueExp:
        # state is its own explanation, no need for annotation
        return Tlacenode(state, None, None, None)
        
    elif type(spec) is FalseExp:    
        print("[ERROR] ARCTL TLACE explain_witness:",
              "cannot explain why state satisfies False",
              spec)
        # TODO Error, cannot explain why state |= False
        return None
    
    elif type(spec) is Atom:
        # state is its own explanation
        return Tlacenode(state, (spec,), None, None)
        
    elif type(spec) is Not:
        return explain_countex(fsm, state, spec.child)
        
    elif type(spec) is And:
        # Get left and right explanations, then merge them
        left = explain_witness(fsm, state, spec.left)
        right = explain_witness(fsm, state, spec.right)
        return Tlacenode(state,
                         left.atomics + right.atomics,
                         left.branches + right.branches,
                         left.universals + right.universals
                        )
        
    elif type(spec) is Or:
        # If state satisfies spec.left, explain it
        # otherwise, state satisfies spec.right, so explain it
        specbdd = evalArctl(fsm, spec.left)
        if state <= specbdd:
            return explain_witness(fsm, state, spec.left)
        else:
            return explain_witness(fsm, state, spec.right)
        
    elif type(spec) is Implies:
        # a -> b is ~a | b
        return explain_witness(fsm, state, Or(Not(spec.left), spec.right))
        
    elif type(spec) is Iff:
        # a <-> b is (a & b) | (~a & ~b)
        return explain_witness(
                fsm, state,
                Or(And(spec.left, spec.right),
                   And(Not(spec.left), Not(spec.right)))
               )
        
    elif type(spec) in {AaF, AaG, AaX, AaU, AaW}:
        # Cannot explain with a single path
        return Tlacenode(state, None, None, (spec,))
                       
    elif type(spec) in {EaF, EaG, EaX, EaU, EaW}:
        branch = explain_branch(fsm, state, spec, spec)
        return Tlacenode(state, None, (branch,), None)
        
    else:
        # TODO Generate error
        print("[ERROR] ARCTL TLACE explain_witness:",
              "unrecognized specification type",
              spec)
        return None
        
        
def explain_countex(fsm, state, spec):
    """
    Explain why state of fsm violates spec.
    
    Return a TLACE node explaining why state of fsm violates spec.
    """
    
    if type(spec) is TrueExp:    
        print("[ERROR] ARCTL TLACE explain_countex:",
              "cannot explain why state violates True",
              spec)
        # TODO Error, cannot explain why state |/= True
        return None
        
    elif type(spec) is FalseExp:
        return explain_witness(fsm, state, TrueExp())
    
    elif type(spec) is Atom:
        # state is its own explanation
        return Tlacenode(state, (Not(spec),), None, None)
        
    elif type(spec) is Not:
        return explain_witness(fsm, state, spec.child)
        
    elif type(spec) is And:
        # ~(a & b) = ~a | ~b
        return explain_witness(fsm, state, Or(Not(spec.left), Not(spec.right)))
        
    elif type(spec) is Or:
        # ~(a | b) = ~a & ~b
        return explain_witness(fsm, state, And(Not(spec.left), Not(spec.right)))
        
    elif type(spec) is Implies:
        # ~(a -> b) = a & ~b
        return explain_witness(fsm, state, And(spec.left, Not(spec.right)))
        
    elif type(spec) is Iff:
        # ~(a <-> b) = (a & ~b) | (~a & b)
        return explain_witness(fsm, state,
                               Or(And(spec.left, Not(spec.right)),
                                  And(Not(spec.left), spec.right)))
        
    elif type(spec) is AaF:
        # ~aaf(a, p) = _eu(a, ~p, ~p & ~_ex(a, true)) | _eg(a, ~p) = eag(a, ~p)
        return explain_witness(fsm, state, EaG(spec.action, Not(spec.child)))
        
    elif type(spec) is AaG:
        return explain_witness(fsm, state, EaF(spec.action, Not(spec.child)))
        
    elif type(spec) is AaX:
        # A<a>X f is false because E<a>X ~f is true or E<a>X true is false
        eaxnf = evalArctl(fsm, EaX(spec.action, Not(spec.child)))
        if state <= eaxnf:
            return explain_witness(fsm, state,
                                   EaX(spec.action, Not(spec.child)))
        else:
            return Tlacenode(state, None, None,
                             (Not(EaX(spec.action, Atom('TRUE'))),))
        
    elif type(spec) is AaU:
        return explain_witness(fsm, state,
                               EaW(spec.action,
                                   Not(spec.right),
                                   And(Not(spec.left), Not(spec.right))))
        
    elif type(spec) is AaW:
        return explain_witness(fsm, state,
                               EaU(spec.action,
                                   Not(spec.right),
                                   And(Not(spec.left), Not(spec.right))))
                     
    elif type(spec) is EaF:
        return explain_witness(fsm, state,
                               AaG(spec.action, Not(spec.child)))
        
    elif type(spec) is EaG:
        return explain_witness(fsm, state,
                               AaF(spec.action, Not(spec.child)))
        
    elif type(spec) is EaX:
        # E<a>X f is false because A<a>X ~f is true or E<a>X true is false
        eaxnf = evalArctl(fsm, AaX(spec.action, Not(spec.child)))
        if state <= eaxnf:
            return explain_witness(fsm, state,
                                   AaX(spec.action, Not(spec.child)))
        else:
            return Tlacenode(state, None, None,
                             (Not(EaX(spec.action, Atom('TRUE'))),))
        
    elif type(spec) is EaU:
        return explain_witness(fsm, state,
                               AaW(spec.action,
                                   Not(spec.right),
                                   And(Not(spec.left), Not(spec.right))))
        
    elif type(spec) is EaW:
        return explain_witness(fsm, state,
                               AaU(spec.action,
                                   Not(spec.right),
                                   And(Not(spec.left), Not(spec.right))))
        
    else:
        # TODO Generate error
        print("[ERROR] ARCTL explain_countex: unrecognized specification type",
              spec)
        return None
        
        
def explain_branch(fsm, state, spec, originalspec):
    """
    Return a TLACE branch explaining why state of fsm satisfies spec.
    """
    if type(spec) is EaX:
        alpha = evalArctl(fsm, spec.action)
        phi = evalArctl(fsm, spec.child)
        path = explain_eax(fsm, state, alpha, phi)
        branch = (Tlacenode(path[0]),
                  path[1],
                  explain_witness(fsm, path[2], spec.child))
        return Tlacebranch(originalspec, branch)

    elif type(spec) is EaF:
        return explain_branch(fsm, state,
                              EaU(spec.action, Atom('TRUE'), spec.child),
                              originalspec)

    elif type(spec) is EaG:
        alpha = evalArctl(fsm, spec.action)
        phi = evalArctl(fsm, spec.child)
        (path, (inloop, loopstate)) = explain_eag(fsm, state, alpha, phi)

        branch = []
        # intermediate states
        loop = None
        for s, i in zip(path[::2], path[1::2]):
            wit = explain_witness(fsm, s, spec.child)
            branch.append(wit)
            branch.append(i)
            # manage the loop
            if loopstate is not None and loopstate == s:
                loop = wit
                
        # last state
        # if loopstate is None, this means that the explanation is a full
        # finite path. We have to explain why this state satisfies ~E<a>X TRUE
        lastnode = explain_witness(fsm, path[-1], spec.child)
        if loopstate is None:
            # Add annotation to show that ~E<a>X TRUE is true
            lastnode = Tlacenode(lastnode.state,
                                 lastnode.atomics,
                                 lastnode.branches,
                                 lastnode.universals +
                                 (Not(EaX(spec.action, Atom('TRUE'))),)
                                )
        branch.append(lastnode)
        
        finalloop = loop and (inloop, loop) or None
        
        return Tlacebranch(originalspec, tuple(branch), finalloop)

    elif type(spec) is EaU:
        alpha = evalArctl(fsm, spec.action)
        phi = evalArctl(fsm, spec.left)
        psi = evalArctl(fsm, spec.right)
        path = explain_eau(fsm, state, alpha, phi, psi)

        branch = []
        # intermediate states
        for s, i in zip(path[::2], path[1::2]):
            branch.append(explain_witness(fsm, s, spec.left))
            branch.append(i)
        # last state
        branch.append(explain_witness(fsm, path[-1], spec.right))

        return Tlacebranch(originalspec, tuple(branch))

    elif type(spec) is EaW:
        eauspec = EaU(spec.action, spec.left, spec.right)
        eagspec = EaG(spec.action, spec.left)
        if state <= evalArctl(fsm, eauspec):
            return explain_branch(fsm, state, eauspec, originalspec)
        else:
            return explain_branch(fsm, state, eagspec, originalspec)

    else:
        # TODO Generate error
        print("[ERROR] ARCTL explain_branch:",
              "unrecognized specification type",
              spec)
        return None


from .tlacenode import Tlacenode
from .tlacebranch import Tlacebranch

from pynusmv_lower_interface.nusmv.parser import parser

from pynusmv.prop import (true as sptrue, false as spfalse, imply, iff,
                               ex, eg, ef, eu, ew, ax, ag, af, au, aw)
from pynusmv.mc import eval_ctl_spec, explainEX, explainEG, explainEU

def explain(fsm, state, spec):
    """
    Return a TLACE node explaining why state of fsm violates spec.
    
    fsm -- a pynusmv.fsm.BddFsm representing the system.
    state -- a pynusmv.dd.BDD representing a state of fsm.
    spec -- a pynusmv.spec.spec.Spec node representing the specification.
    
    Return a tlacenode.Tlacenode explaining why state of fsm violates spec.
    """
    return countex(fsm, state, spec, None)
    
    
def countex(fsm, state, spec, context):
    """
    Return a TLACE node explaining why state of fsm violates spec.
    
    fsm -- a pynusmv.fsm.BddFsm representing the system.
    state -- a pynusmv.dd.BDD representing a state of fsm.
    spec -- a pynusmv.spec.spec.Spec node representing the specification.
    context -- a pynusmv.spec.spec.Spec representing the context of spec in fsm.
    
    Return a tlacenode.Tlacenode explaining why state of fsm violates spec.
    """
    
    if spec.type == parser.CONTEXT:
        return countex(fsm, state, spec.cdr, spec.car)
        
    elif spec.type == parser.FALSEEXP:
        newspec = sptrue()
        
    elif spec.type == parser.NOT:
        newspec = spec.car
        
    elif spec.type == parser.OR:
        newspec = (~spec.car) & (~spec.cdr)
    
    elif spec.type == parser.AND:
        newspec = (~spec.car) | (~spec.cdr)
    
    elif spec.type == parser.IMPLIES:
        newspec = spec.car & (~spec.cdr)
                            
    elif spec.type == parser.IFF:
        newspec = (spec.car & (~spec.cdr)) | ((~spec.car) & spec.cdr)
                    
    elif spec.type == parser.EX:
        newspec = ax(~spec.car)
                                 
    elif spec.type == parser.EF:
        newspec = ag(~spec.car)
                                 
    elif spec.type == parser.EG:
        newspec = af(~spec.car)
                                 
    elif spec.type == parser.EU:
        newspec = aw(~spec.cdr, (~spec.car) & (~spec.cdr))
                    
    elif spec.type == parser.EW:
        newspec = au(~spec.cdr, (~spec.car) & (~spec.cdr))
                    
    elif spec.type == parser.AX:
        newspec = ex(~spec.car)
        
    elif spec.type == parser.AF:
        newspec = eg(~spec.car)
                                 
    elif spec.type == parser.AG:
        newspec = ef(~spec.car)
                                 
    elif spec.type == parser.AU:
        newspec = ew(~spec.cdr, (~spec.car) & (~spec.cdr))
                        
    elif spec.type == parser.AW:
        newspec = eu(~spec.cdr, (~spec.car) & (~spec.cdr))
                        
    else:
        if spec.type == parser.NOT:
            newspec = spec.car
        else:
            newspec = ~spec
        return Tlacenode(state, (newspec,), None, None)
        
    return witness(fsm, state, newspec, context)
    
    
def witness(fsm, state, spec, context):
    """
    Return a TLACE node explaining why state of fsm satisfies spec.
    
    fsm -- a pynusmv.fsm.BddFsm representing the system.
    state -- a pynusmv.dd.BDD representing a state of fsm.
    spec -- a pynusmv.spec.spec.Spec node representing the specification.
    context -- a pynusmv.spec.spec.Spec representing the context of spec in fsm.
    
    Return a tlacenode.Tlacenode explaining why state of fsm satisfies spec.
    """

    if spec.type == parser.CONTEXT:
        return witness(fsm, state, spec.cdr, spec.car)
        
    elif spec.type == parser.TRUEEXP:
        return Tlacenode(state, None, None, None)
        
    elif spec.type == parser.NOT:
        return countex(fsm, state, spec.car, context)
        
    elif spec.type == parser.OR:
        if state.entailed(eval_ctl_spec(fsm, spec.car, context)):
            return witness(fsm, state, spec.car, context)
        else:
            return witness(fsm, state, spec.cdr, context)
    
    elif spec.type == parser.AND:
        n1 = witness(fsm, state, spec.car, context)
        n2 = witness(fsm, state, spec.cdr, context)
        return Tlacenode(state,
                         n1.atomics + n2.atomics,
                         n1.branches + n2.branches,
                         n1.universals + n2.universals)
    
    elif spec.type == parser.IMPLIES:
        newspec = (~spec.car) | spec.cdr
        return witness(fsm, state, newspec, context)
                            
    elif spec.type == parser.IFF:
        newspec = (spec.car & spec.cdr) | ((~spec.car) & (~spec.cdr))
        return witness(fsm, state, newspec, context)
                    
    elif (spec.type == parser.EX or
          spec.type == parser.EF or
          spec.type == parser.EG or
          spec.type == parser.EU or
          spec.type == parser.EW):
        return Tlacenode(state, None,
                         (witness_branch(fsm, state, spec, context, spec),),
                         None)
                    
    elif (spec.type == parser.AX or
          spec.type == parser.AF or
          spec.type == parser.AG or
          spec.type == parser.AU or
          spec.type == parser.AW):
        return Tlacenode(state, None, None, (spec,))
                        
    else:
        # deal with unmanaged formulas by returning a single
        # TLACE node. This includes the case of atomic propositions.
        # All unrecognized operators are considered as atomics
        return Tlacenode(state, (spec,), None, None)
        
        
def witness_branch(fsm, state, spec, context, originalspec):
    """
    Return a TLACE branch explaining why state of fsm satisfies spec.

    fsm -- a pynusmv.fsm.BddFsm representing the system.
    state -- a pynusmv.dd.BDD representing a state of fsm.
    spec -- a pynusmv.spec.spec.Spec node representing the specification.
    context -- a pynusmv.spec.spec.Spec representing the context of spec in fsm.
    originalspec -- a pynusmv.spec.spec.Spec representing the original spec;
                    used to annotate the produced branch, despite updated
                    specs.

    Return a tlacebranch.Tlacebranch explaining why state of fsm satisfies spec.
    
    Throw a NonExistentialSpecError if spec is not existential.
    """
    
    if spec.type == parser.EX:
        f = eval_ctl_spec(fsm, spec.car, context)
        path = explainEX(fsm, state, f)
        branch = (Tlacenode(path[0]),
                  path[1],
                  witness(fsm, path[2], spec.car, context))
        return Tlacebranch(originalspec, branch)
        
    elif spec.type == parser.EF:
        newspec = eu(sptrue(), spec.car)
        return witness_branch(fsm, state, newspec, context, originalspec)
        
    elif spec.type == parser.EG:
        f = eval_ctl_spec(fsm, spec.car, context)
        (path, (inloop, loopstate)) = explainEG(fsm, state, f)
        
        branch = []
        # intermediate states
        for s, i in zip(path[::2], path[1::2]):
            wit = witness(fsm, s, spec.car, context)
            branch.append(wit)
            branch.append(i)
            # manage the loop
            if s == loopstate:
                loop = wit
        # last state
        branch.append(witness(fsm, path[-1], spec.car, context))
        
        return Tlacebranch(originalspec, tuple(branch), (inloop, loop))
        
    elif spec.type == parser.EU:
        f = eval_ctl_spec(fsm, spec.car, context)
        g = eval_ctl_spec(fsm, spec.cdr, context)
        path = explainEU(fsm, state, f, g)
        
        branch = []
        # intermediate states
        for s, i in zip(path[::2], path[1::2]):
            branch.append(witness(fsm, s, spec.car, context))
            branch.append(i)
        # last state
        branch.append(witness(fsm, path[-1], spec.cdr, context))
        
        return Tlacebranch(originalspec, tuple(branch))
        
    elif spec.type == parser.EW:
        euspec = eu(spec.car, spec.cdr)
        egspec = eg(spec.car)
        if state.entailed(eval_ctl_spec(fsm, euspec, context)):
            return witness_branch(fsm, state, euspec, context, originalspec)
        else:
            return witness_branch(fsm, state, egspec, context, originalspec)
        
    else:
        # Default case, throw an exception because spec is not existential
        raise NonExistentialSpecError()



class NonExistentialSpecError(Exception):
    """
    Exception for given non existential temporal formula.
    """
    pass



class Tlacenode(object):
    """
    A Tlacenode is a TLACE node.
    
    It contains a state, a list of atomic propositions, a list of TLACE branches
    and a list of universal formulas.
    """
    
    def __init__(self, state, atomics=None, branches=None, universals=None):
        """
        Create a new TLACE node.
        
        state -- a State representing the state of the TLACE node
        atomics -- a list of atomic propositions,
                   represented by CTLK AST instances
        branches -- a list of TLACE branches represented by Tlacebranches
        universals -- a list of universal formulas,
                      represented by CTLK AST instances
        """
        self._state = state
        self._atomics = atomics or tuple()
        self._branches = branches or tuple()
        self._universals = universals or tuple()
    
    @property
    def state(self):
        """state node"""
        return self._state
        
    @property
    def atomics(self):
        """atomic annotations of this node"""
        return self._atomics
        
    @property
    def branches(self):
        """branches of this node"""
        return self._branches
        
    @property
    def universals(self):
        """universal annotations of this node"""
        return self._universals


class Tlacebranch(object):
    """
    A Tlacebranch is a TLACE branch.
    It contains an existential formula.
    This is a generic TLACE branch that must be extended
    to provide an explanation to the existential formula.
    """
    
    def __init__(self, formula):
        """
        Create a new TLACE branch.
        
        formula -- an existential temporal formula,
                   represented by a CTLK AST instance.
        """
        self._formula = formula      
    
    @property
    def specification(self):
        """The spec of this branch."""
        return self._formula
        
        
class TemporalBranch(Tlacebranch):
    """A TLACE branch explaining an existential temporal property."""
    
    def __init__(self, formula, path, loop=None):
        """
        Create a new existential temporal branch.
        
        formula -- the formula this branch explains, as a CTLK AST instance
        path -- the finite path of the explanation,
                as a (state, input, state, ..., input, state) tuple
        loop -- if not None, an (input, state) pair where state belongs to path,
                indicating the start of the loop; input being an input between
                path[-1] and state.
        """
        super(TemporalBranch, self).__init__(formula)
        self._path = path
        self._loop = loop
        
    @property
    def path(self):
        """The path of this branch"""
        return self._path
        
    @property
    def loop(self):
        """The loop of this branch, possibly None"""
        return self._loop
    
    
class EpistemicBranch(Tlacebranch):
    """A TLACE branch explaining an existential epistemic property."""
    
    def __init__(self, formula, path):
        """
        Create a new existential epistemic branch.
        
        formula -- the formula this branch explains, as a CTLK AST instance
        path -- the finite path of the explanation,
                as a (s_0, ag_1, s_1, ..., ag_n, state_n) tuple,
                where ag_i is the name of the agents
                for which s_i-1 is equivalent to s_i, for all i : 0 < i <= n.
        """
        super(EpistemicBranch, self).__init__(formula)
        self._path = path
        
    @property
    def path(self):
        """The path of this branch"""
        return self._path


#! /usr/local/bin/python3
"""
This is an alternative implementation which uses the native C api as they were
brought by the the ltlspec module.

Concretely it uses the apis defined in the `pynusmv_lower_interface.bmc_utils`
part of the lower interface which extends the ones exposed by nusmv. Hence, this
script does the same as `alternative1` but decomposes the problem generation in

    * Generating the model formula
    * Generating a path formula of length k
    * Generating the property formula

.. and combine all those bits together.
"""
import sys
import argparse

from pynusmv.init         import init_nusmv
from pynusmv.glob         import load
from pynusmv.bmc.glob     import BmcSupport, master_be_fsm
from pynusmv.parser       import parse_ltl_spec
from pynusmv.node         import Node
from pynusmv.sat          import SatSolverFactory, Polarity, SatSolverResult
from pynusmv.bmc          import ltlspec, utils as bmcutils

def arguments():
    """
    Creates the arguments parser and manages to react to wrong usage.

    :returns: an object having field to store each of the command line arguments
    """
    parser = argparse.ArgumentParser(description="a PyNuSMV backed LTL sat based bmc verifier for LTL")
    parser.add_argument("-v", "--verbose", action="store_true", help="Displays the text of the analyzed model")
    parser.add_argument("-k", "--bound",   type=int, default=10, help="the maximum number of steps in a verified path")
    parser.add_argument("-s", "--spec",    type=str, help="the LTL specification to verify")
    parser.add_argument("-f", "--no-fairness",  help="disable the use of fairness constraints", action="store_true")
    parser.add_argument("-i", "--no-invariants",help="disable the invariants enforcement", action="store_true")
    parser.add_argument("-d", "--dry-run", action="store_true", help="do not perform the verification (no sat solving)")
    parser.add_argument("model", type=str, help="the name of a file containing an SMV model")

    return parser.parse_args()

def check_problem(pb, length):
    fsm    = master_be_fsm()
    cnf    = pb.to_cnf(Polarity.POSITIVE)

    solver = SatSolverFactory.create()
    solver+= cnf
    solver.polarity(cnf, Polarity.POSITIVE)

    if solver.solve() == SatSolverResult.SATISFIABLE:
        cnt_ex = bmcutils.generate_counter_example(fsm, pb, solver, length, "Violation")
        return ("Violation", cnt_ex)
    else:
        return ("Ok", None)


def generate_path(offset, length):
    """
    Returns a boolean expression representing a path of length `length` in the
    fsm described by the loaded model.

    :param length: the length of the path in the fsm
    :param offset: the offset at which the path should be starting
    :return: a boolean expression representing a path of length `length` in the
        loaded fsm.
    """
    model = bmcutils.BmcModel()
    path  = model.init[offset] & model.unrolling(offset, offset + length)
    return path

def generate_sat_problem(fsm, fml, length):
    fsm = master_be_fsm()
    offset  = 0
    wff     = bmcutils.make_negated_nnf_boolean_wff(fml).to_node()
    problem = generate_path(offset, length) \
            & ltlspec.bounded_semantics_at_offset(fsm, wff, length, offset)
    return problem

def check_ltl(fml, bound, dry_run):
    fsm     = master_be_fsm()

    import time
    for i in range(bound+1):
        start = time.time()
        problem = generate_sat_problem(fsm, fml, i)
        end   = time.time()
        if not dry_run:
            status, trace = check_problem(problem, i)
            if status != "Ok":
                return (status, i, trace)
            else:
                print("-- No problem at length {}".format(i))
        else:
            print(" 'Problem {}' ; {}".format(i, end-start))

    return ("Ok", bound, None)

def check(formula, args):
    parsed_fml          = Node.from_ptr(parse_ltl_spec(formula.strip()))
    status,length,trace = check_ltl(parsed_fml, args.bound, args.dry_run)
    if status != 'Ok':
        print("-- {} for length {}".format(status, length))
        print(trace)


def main():
    args = arguments()

    with init_nusmv():
        load(args.model)
        if args.verbose:
            with open(args.model) as f:
                print(f.read())

        with BmcSupport():
            if args.spec is not None:
                check(args.spec, args)
            else:
                print("Enter LTL properties, one per line:")
                for line in sys.stdin:
                    check(line, args)

if __name__ == "__main__":
    main()



class Tlacenode:
    """
    A Tlacenode is a TLACE node.
    
    It contains a state, a list of atomic propositions, a list of TLACE branches
    and a list of universal formulas.
    """
    
    def __init__(self, state, atomics=None, branches=None, universals=None):
        """
        Create a new TLACE node.
        
        state -- a BDD representing the state of the TLACE node
        atomics -- a list of atomic propositions represented by Nodes
        branches -- a list of TLACE branches represented by Tlacebranches
        universals -- a list of universal formulas, represented by Nodes
        """
        self.__state = state
        self.__atomics = atomics or tuple()
        self.__branches = branches or tuple()
        self.__universals = universals or tuple()
    
    @property
    def state(self):
        """state node"""
        return self.__state
        
    @property
    def atomics(self):
        """atomic annotations of this node"""
        return self.__atomics
        
    @property
    def branches(self):
        """branches of this node"""
        return self.__branches
        
    @property
    def universals(self):
        """universal annotations of this node"""
        return self.__universals
        
    def __str__(self):
        pass # TODO

class Tlacebranch:
    """
    A Tlacebranch is a TLACE branch.
    It contains an existential formula and a TLACE path,
    represented by a tuple of a list of TLACE nodes
    and a possibly None looping node.
    """
    
    def __init__(self, formula, path, loop=None):
        """
        Creates a new TLACE branch.
        formula is an existential temporal formula;
        path is a list of TLACE nodes;
        loop is None or a (inputs, node) pair where node is a node of path
        indicating the start of a loop.
        """
        self.__formula = formula
        self.__path = (path, loop)        
    
    @property
    def specification(self):
        """The spec of this branch."""
        return self.__formula
        
    @property
    def path(self):
        """The path of this branch. A (path, (inputs, loop)) tuple."""
        return self.__path
        
    def __str__(self):
        pass # TODO


from pynusmv.dd import BDD
from pynusmv.fsm import BddFsm, BddTrans
from pynusmv_lower_interface.nusmv.node import node as nsnode
from pynusmv_lower_interface.nusmv.parser import parser as nsparser
from pynusmv_lower_interface.nusmv.dd import dd as nsdd
from .exception import UnknownAgentError

class MAS(BddFsm):
    """
    A multi-agent system.
    
    A multi-agent system is an FSM with one temporal trans for the whole model
    and one epistemic trans by agent, labelled with the name of the agent.
    
    epistemic -- a dictionary of agent->the NuSMV node-based TRANS of the agent.
    """
    
    def __init__(self, ptr, observed, inputvars, epistemic, groups=None,
                 freeit=False):
        """
        Create a new MAS.
        
        ptr -- the pointer of the NuSMV FSM
        observed -- a dictionary of agents' known variables
                    (agent->Set of vars)
                    representing the variables observed by each agent
        inputvars -- a dictionary of agents' input variables
                     (agent->Set of input vars)
        epistemic -- a dictionary of epistemic relations (agent->EPISTEMIC)
        groups -- if not None, a dictionary of group name -> set of names of
                  the agents of the group; the groups names and the agents
                  names must be present in epistemic
        freeit -- whether or not free the pointer
        """
        super(MAS, self).__init__(ptr, freeit=freeit)
        self._epistemic = epistemic and epistemic or {}
        self._epistemic_trans = {}
        self._agents_observed_variables = observed
        self._agents_inputvars = inputvars
        self._groups = groups if groups is not None else {}
        self._protocols = {}
        
    
    @property
    def agents(self):
        """The list of names of the agents of the system."""
        return set(self._epistemic.keys())
        
    
    @property
    def agents_observed_variables(self):
        """The state variables observed by each agent (as a dictionary)."""
        return self._agents_observed_variables
        
    
    @property
    def agents_inputvars(self):
        """The input variables of each agent (as a dictionary)."""
        return self._agents_inputvars
    
    @property
    def groups(self):
        """The set of groups of the MAS, with corresponding agents."""
        return self._groups
        
        
    def _compute_epistemic_trans(self, agents):
        from .glob import symb_table
        trans = None
        for agent in agents:
            if agent not in self._epistemic:
                raise UnknownAgentError(str(agents) +
                                        " are an unknown agents names.")
            trans = nsnode.find_node(nsparser.AND,
                                     self._epistemic[agent],
                                     trans)
        self._epistemic_trans[frozenset(agents)] = BddTrans.from_trans(
                                                            symb_table(),
                                                            trans)
    
    def equivalent_states(self, states, agents):
        """
        Return the BDD representing the set of states epistemically equivalent
        to states through the epistemic relation of agents. agents is a set of
        agents name.
        
        states -- a BDD of states of self
        agents -- a set of agents names.
                  These agents must be ones of this MAS agents.
        """
            
        # Apply FSM constraints
        states = states & self.state_constraints
        
        # Compute the post-image
        if frozenset(agents) not in self._epistemic_trans:
            # Compute the BddTrans
            self._compute_epistemic_trans(agents)
        # pre_ or post_ operations can be used since the epistemic relations
        # are equivalence relations and thus are symetric
        result = self._epistemic_trans[frozenset(agents)].pre(states)
        
        # Apply constraints on result
        return result & self.state_constraints
        
    
    def protocol(self, agents):
        """
        Return the protocol for the given set of agents.
        
        agents -- a set of agent names
        
        The returned value is a BDD representing state-input pairs, such that
        for every state, the given inputs (for agents only) are the ones
        available in the state, i.e. for a state s, <s,a_gamma> is
        in the returned BDD iff there exist a_ngamma (an action for all other
        agents) and s' such that <s, a_gamma, a_ngamma, s'> is in the transition
        relation.
        
        """
        agents = frozenset(agents)
        if agents not in self._protocols:
            gamma_inputs = [var
                            for agent in agents
                            for var in self.agents_inputvars[agent]]
            gamma_cube = self.bddEnc.cube_for_inputs_vars(gamma_inputs)
            ngamma_cube = self.bddEnc.inputsCube - gamma_cube
            self._protocols[agents] = (self.weak_pre(self.reachable_states).
                                       forsome(ngamma_cube) &
                                       self.reachable_states &
                                       self.bddEnc.statesInputsMask)
                                       
        return self._protocols[agents]
        
        
    def inputs_cube_for_agents(self, agents):
        """
        Return the cube of inputs variables of agents.
        
        agents -- a set of agents names of this MAS.
        
        """
        gamma_inputs = [var
                        for agent in agents
                        for var in self.agents_inputvars[agent]]
        return self.bddEnc.cube_for_inputs_vars(gamma_inputs)
        
    def pre(self, states, inputs=None, subsystem=None):
        """
        Return the pre image of states, through inputs (if any) and in
        subsystem (if any).
        """
        
        if inputs is None:
            inputs = BDD.true(self.bddEnc.DDmanager)
        
        if subsystem is None:
            subsystem = BDD.true(self.bddEnc.DDmanager)
        
        return ((self.weak_pre(states & inputs) &
                 subsystem).forsome(self.bddEnc.inputsCube) &
                 self.bddEnc.statesMask)
    
    def post(self, states, inputs=None, subsystem=None):
        """
        Return the post image of states, through inputs (if any) and in
        subsystem (if any).
        """
        
        if inputs is None:
            inputs = BDD.true(self.bddEnc.DDmanager)
        
        if subsystem is None:
            subsystem = BDD.true(self.bddEnc.DDmanager)
        
        states = states & subsystem
        
        return super(MAS, self).post(states, inputs)
        
        
    def pre_strat(self, states, agents, strat=None):
        """
        Return the set of states s of this MAS such that there exists values
        of input variables of the agents such that for all values of input
        variables of the other agents, all successors of s through these inputs
        belong to states.
        If strat is not None, restrict to strat.
        
        states -- a BDD representing a set of states of this MAS;
                  if states represents a set of state/inputs pairs, inputs
                  are abstracted away;
        agents -- a set of agents names, agents of this MAS;
        strat -- a BDD representing allowed state/inputs pairs, or None.
        
        """
        if not strat:
            strat = BDD.true(self.bddEnc.DDmanager)
            
        gamma_cube = self.inputs_cube_for_agents(agents)
        ngamma_cube = self.bddEnc.inputsCube - gamma_cube
        
        # Abstract away actions of states
        states = states.forsome(self.bddEnc.inputsCube)
        
        nstates = ~states & self.bddEnc.statesInputsMask
        strat = strat & self.bddEnc.statesInputsMask
        
        return (
                ( ~(
                    self.weak_pre(nstates).forsome(ngamma_cube)
                  )
                  & 
                  self.weak_pre(states)
                ).forsome(ngamma_cube)
                &
                strat
                &
                self.bddEnc.statesInputsMask
               ).forsome(self.bddEnc.inputsCube)
                
                
    def pre_nstrat(self, states, agents):
        """
        Return the set of states s of this MAS such that for all values of input
        variables of the agents there exists values of input
        variables of the other agents such that there is a successor of s
        through these inputs that belongs to states.
        
        states -- a BDD representing a set of states of this MAS;
        agents -- a set of agents names, agents of this MAS.
        
        """
        return ~self.pre_strat(~states, agents)
        
    
    def pre_strat_si(self, states, agents, strat=None):
        """
        Return the set of state/inputs pairs <s,i_agents> of this MAS such that
        there exist values of input variables of the agents i_agents,
        all successors of s through i_agents U i_nagents belong to states.
        Restrict to strat if not None.
        
        states -- a BDD representing a set of states of this MAS;
                  if states represents a set of state/inputs pairs, inputs
                  are abstracted away;
        agents -- a set of agents names, agents of this MAS;
        strat -- a BDD representing a set of allowed state/inputs pairs.
        
        """
        if strat is None:
            strat = BDD.true(self.bddEnc.DDmanager)
        
        gamma_cube = self.inputs_cube_for_agents(agents)
        ngamma_cube = self.bddEnc.inputsCube - gamma_cube
        
        # Abstract inputs from states to avoid mixing with the possible actions
        # present in states.
        states = states.forsome(self.bddEnc.inputsCube)
        
        nstates = ~states & self.bddEnc.statesInputsMask
        strat = strat & self.bddEnc.statesInputsMask
                
        return (
                ( ~(
                    self.weak_pre(nstates).forsome(ngamma_cube)
                  )
                  & 
                  self.weak_pre(states)
                ).forsome(ngamma_cube)
                &
                strat
                &
                self.bddEnc.statesInputsMask
               )
    
    def check_free_choice(self):
        """
        Check whether this MAS satisfies the free-choice property, that is,
        in every state, the choices of actions for each agent is not
        constrained by the choices of other agents.
        
        Return the set of moves that are not present in the MAS and should,
        or that are present but should not.
        """
        
        if len(self.agents) <= 0:
            return BDD.false(self.bddEnc.DDmanager)
        
        true = BDD.true(self.bddEnc.DDmanager)
        protocols = {agent: self.protocol({agent}) for agent in self.agents}
        enabled = (self.weak_pre(self.reachable_states) &
                   self.reachable_states & self.bddEnc.statesInputsMask)
        
        for s in self.pick_all_states(self.reachable_states):
            product = self.bddEnc.statesInputsMask
            for agent in self.agents:
                product &= protocols[agent] & s
            if (enabled & s) != product:
                return product.xor(enabled & s)
        return BDD.false(self.bddEnc.DDmanager)
    
    def check_uniform_choice(self):
        """
        Check whether this MAS satisfies the uniform-choice property, that is,
        the action of each agent in two epistemically equivalent states are
        the same.
        
        Return a FALSE BDD if this MAS satisfies the uniform-choice property,
        or a set of moves showing why this MAS does not satisfy it, otherwise.
        """
        
        true = BDD.true(self.bddEnc.DDmanager)
        false = BDD.false(self.bddEnc.DDmanager)
        inputs_cube = self.bddEnc.inputsCube
        states_cube = self.bddEnc.statesCube
        
        if len(self.agents) <= 0:
            return false
        
        for agent in self.agents:
            act_cube = self.inputs_cube_for_agents({agent})
            protocol = self.protocol({agent})
            remaining = protocol
            
            while remaining.isnot_false():
                # One state
                s = self.pick_one_state(remaining.forsome(inputs_cube))
                # The equivalence class of s
                eq = (self.equivalent_states(s, {agent}) &
                      self.reachable_states)
                # The actions for agent in s
                i = (protocol & s).forsome(states_cube)
                
                for sp in self.pick_all_states(eq):
                    ip = (protocol & sp).forsome(states_cube)
                    if ip != i:
                        return (s & i) | (sp & ip)
                
                remaining = remaining - eq
        return false
    
    def check_mas(self):
        """
        Check whether this MAS satisfies the free-choice and uniform-choice
        properties, that is,
        - in every state, the choices of actions for each agent is not
          constrained by the choices of other agents;
        - the action of each agent in two epistemically equivalent states are
          the same.
        """
        return (self.check_free_choice().is_false() and
                self.check_uniform_choice().is_false())
        


class Agent(object):
    """
    An agent has a name, a set of observable variables and a set of actions.
    Observable variables must be NuSMV variables (no define, frozen variable
    or input variable).
    Actions must be NuSMV input variables.
    """
    def __init__(self, name, observables, actions):
        self.name = name
        self.observables = observables
        self.actions = actions

class Group(Agent):
    """
    A group is an agent representing a group of agents.
    """
    def __init__(self, name, *agents):
        self.name = name
        self.agents = agents
        self.observables = set(observable for agent in agents
                                          for observable in agent.observables)
        self.actions = set(action for agent in agents
                                  for action in agent.actions)


import sys
import argparse
from pynusmv.init import init_nusmv, deinit_nusmv
from ..mas import glob
from .parsing import parseATL
from .eval import evalATL
from pyparsing import ParseException
from pynusmv.exception import PyNuSMVError

def check(mas, spec):
    """
    Return whether the system satisfies the ATL specification.
    
    mas -- the system
    spec -- the specification, as a string
    
    """
    sat = evalATL(mas, spec)
    return (~sat & mas.bddEnc.statesInputsMask & mas.init).is_false()    
    
def process(allargs):
    """
    Check specs on the given NuSMV model.
    
    Build the model from a given file and check every ATL specification
    writen on standard input.
    
    allargs -- a sys.args-like arguments list, without script name.
    """
    
    # Parse arguments
    parser = argparse.ArgumentParser(description='ATL model checker.')
    # Populate arguments: for now, only the model
    parser.add_argument('model', help='the MAS as an SMV model')
    parser.add_argument('-p', dest='property', help='the property check',
                        default=None)
    args = parser.parse_args(allargs)
    
    # Initialize the model
    glob.load_from_file(args.model)
    mas = glob.mas()
    
    # Check given property, if any
    if args.property:
        try:
            spec = parseATL(args.property)[0]
            satisfied = check(mas, spec)
            print('Specification', str(spec), 'is', str(satisfied))
        except ParseException as e:
            print("[ERROR] Cannot parse specification:", str(e))
        except PyNuSMVError as e:
            print("[ERROR]", str(e))
    
    # Otherwise, ask for ATL properties
    else:
        print("Enter ATL properties, one by line:")
        for line in sys.stdin:
            try:
                spec = parseATL(line)[0]
            
                satisfied = check(mas, spec)
                print('Specification', str(spec), 'is', str(satisfied))
            except ParseException as e:
                print("[ERROR] Cannot parse specification:", str(e))
            except PyNuSMVError as e:
                print("[ERROR]", str(e))

def main():
    with init_nusmv():
        process(sys.argv[1:])

if __name__ == '__main__':
    main()



import unittest

from pyparsing import ParseException

from pynusmv_tools.arctl.parsing import parseArctl
from pynusmv_tools.arctl.ast import (Atom, Not, And, Or, Implies, Iff, 
                             AaF, AaG, AaX, AaU, AaW,
                             EaF, EaG, EaX, EaU, EaW)

class TestParsing(unittest.TestCase):
    
    def test_atom(self):
        s = "'c <= 3'"
        asts = parseArctl(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertEqual(type(ast), Atom)
        self.assertEqual(ast.value, "c <= 3")
        
    
    def test_fail(self):
        kos = ["", "'test", "A<'test'>H 'q'", "O", "A<'a'>E<'e'>G 'true'",
               "E<'ac'>[A<'ac'>['a' E<'ac'>['b' W 'c'] W 'd'] "
               "U A<'ac'>['e' U 'f']]", "A<'t't'>G 'q'"]
        
        for s in kos:
            with self.assertRaises(ParseException):
                parseArctl(s)
    
    
    def test_not(self):
        s = "~'a'"
        asts = parseArctl(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertEqual(type(ast), Not)
        self.assertEqual(type(ast.child), Atom)
        self.assertEqual(ast.child.value, "a")
        
    
    def test_and(self):
        s = "'a' & ('b' & 'c')"
        asts = parseArctl(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertEqual(type(ast), And)
        self.assertEqual(type(ast.left), Atom)
        self.assertEqual(ast.left.value, "a")
        self.assertEqual(type(ast.right), And)
        self.assertEqual(type(ast.right.left), Atom)
        self.assertEqual(ast.right.left.value, "b")
        self.assertEqual(type(ast.right.right), Atom)
        self.assertEqual(ast.right.right.value, "c")
        
    
    def test_implies(self):
        s = "'a' -> 'b'"
        asts = parseArctl(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertEqual(type(ast), Implies)
        self.assertEqual(type(ast.left), Atom)
        self.assertEqual(ast.left.value, "a")
        self.assertEqual(type(ast.right), Atom)
        self.assertEqual(ast.right.value, "b")
        
        
    def test_logicals(self):
        s = "'a' & 'c' -> (~'b' | 'c')"
        asts = parseArctl(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertEqual(type(ast), Implies)
        self.assertEqual(type(ast.left), And)
        self.assertEqual(type(ast.left.left), Atom)
        self.assertEqual(ast.left.left.value, "a")
        self.assertEqual(type(ast.left.right), Atom)
        self.assertEqual(ast.left.right.value, "c")
        self.assertEqual(type(ast.right), Or)
        self.assertEqual(type(ast.right.left), Not)
        self.assertEqual(type(ast.right.left.child), Atom)
        self.assertEqual(ast.right.left.child.value, "b")
        self.assertEqual(type(ast.right.right), Atom)
        self.assertEqual(ast.right.right.value, "c")
        
    
    def test_actions(self):
        s = "A<'a' & 'b' | 'c'>F 'b'"
        asts = parseArctl(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertEqual(type(ast), AaF)
        self.assertEqual(type(ast.action), Or)
        self.assertEqual(type(ast.action.left), And)
        self.assertEqual(type(ast.action.left.left), Atom)
        self.assertEqual(ast.action.left.left.value, "a")
        self.assertEqual(type(ast.action.left.right), Atom)
        self.assertEqual(ast.action.left.right.value, "b")
        self.assertEqual(type(ast.action.right), Atom)
        self.assertEqual(ast.action.right.value, "c")
        
        self.assertEqual(type(ast.child), Atom)
        self.assertEqual(ast.child.value, "b")
        
        
    def test_x(self):
        s = "E<'a'>X (A<'b'> X 'c' & E<'f'>X E<'g'>X 'h')"
        asts = parseArctl(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertEqual(type(ast), EaX)
        self.assertEqual(type(ast.action), Atom)
        self.assertEqual(ast.action.value, "a")
        self.assertEqual(type(ast.child), And)
        self.assertEqual(type(ast.child.left), AaX)
        self.assertEqual(type(ast.child.left.action), Atom)
        self.assertEqual(ast.child.left.action.value, "b")
        self.assertEqual(type(ast.child.left.child), Atom)
        self.assertEqual(ast.child.left.child.value, "c")
        self.assertEqual(type(ast.child.right), EaX)
        self.assertEqual(type(ast.child.right.action), Atom)
        self.assertEqual(ast.child.right.action.value, "f")
        self.assertEqual(type(ast.child.right.child), EaX)
        self.assertEqual(type(ast.child.right.child.action), Atom)
        self.assertEqual(ast.child.right.child.action.value, "g")
        self.assertEqual(type(ast.child.right.child.child), Atom)
        self.assertEqual(ast.child.right.child.child.value, "h")
        
        
    def test_ef_and(self):
        s = "E<'a'>F 'admin = alice' & E<'b'>F 'admin = bob'"
        asts = parseArctl(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertEqual(type(ast), And)
        self.assertEqual(type(ast.left), EaF)
        self.assertEqual(type(ast.left.action), Atom)
        self.assertEqual(ast.left.action.value, "a")
        self.assertEqual(type(ast.left.child), Atom)
        self.assertEqual(ast.left.child.value, "admin = alice")
        self.assertEqual(type(ast.right), EaF)
        self.assertEqual(type(ast.right.action), Atom)
        self.assertEqual(ast.right.action.value, "b")
        self.assertEqual(type(ast.right.child), Atom)
        self.assertEqual(ast.right.child.value, "admin = bob")
        
        
    def test_not_eax(self):
        s = "E<'a'>X ~ 'c'"
        asts = parseArctl(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertEqual(type(ast), EaX)
        self.assertEqual(type(ast.action), Atom)
        self.assertEqual(ast.action.value, "a")
        self.assertEqual(type(ast.child), Not)
        self.assertEqual(type(ast.child.child), Atom)
        self.assertEqual(ast.child.child.value, "c")
        
        
    def test_afnext(self):
        s = "A<'TRUE'>F (~ E<'TRUE'>X 'TRUE')"
        asts = parseArctl(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertEqual(type(ast), AaF)
        self.assertEqual(type(ast.action), Atom)
        self.assertEqual(ast.action.value, "TRUE")
        self.assertEqual(type(ast.child), Not)
        self.assertEqual(type(ast.child.child), EaX)
        self.assertEqual(type(ast.child.child.action), Atom)
        self.assertEqual(ast.child.child.action.value, "TRUE")
        self.assertEqual(type(ast.child.child.child), Atom)
        self.assertEqual(ast.child.child.child.value, "TRUE")
        
    
    def test_eauw(self):
        s = "E<'ac'>[A<'ac'>[E<'ac'>['b' W 'c'] W 'd'] U A<'ac'>['e' U 'f']]"
        asts = parseArctl(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertEqual(type(ast), EaU)
        self.assertEqual(type(ast.action), Atom)
        self.assertEqual(ast.action.value, "ac")
        self.assertEqual(type(ast.left), AaW)
        self.assertEqual(type(ast.left.action), Atom)
        self.assertEqual(ast.left.action.value, "ac")
        self.assertEqual(type(ast.left.right), Atom)
        self.assertEqual(ast.left.right.value, "d")
    
    
    def test_full(self):
        s = ("A<'past'>G (E<'true'>F 'future is now'"
             "<-> A<'min'>['past' U 'present'])")
        asts = parseArctl(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertEqual(type(ast), AaG)
        self.assertEqual(type(ast.action), Atom)
        self.assertEqual(ast.action.value, "past")
        self.assertEqual(type(ast.child), Iff)
        self.assertEqual(type(ast.child.left), EaF)
        self.assertEqual(type(ast.child.left.action), Atom)
        self.assertEqual(ast.child.left.action.value, "true")
        self.assertEqual(type(ast.child.left.child), Atom)
        self.assertEqual(ast.child.left.child.value, "future is now")
        self.assertEqual(type(ast.child.right), AaU)
      
        
    def test_full2(self):
        s = ("A<'past' | 'present'>G (E<'true'>F ('future is now' & 'later')"
             "<-> A<'min'>['past' U 'present' -> 'future'])")
        asts = parseArctl(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertEqual(type(ast), AaG)
        self.assertEqual(type(ast.action), Or)
        self.assertEqual(type(ast.action.left), Atom)
        self.assertEqual(ast.action.left.value, "past")
        self.assertEqual(type(ast.action.right), Atom)
        self.assertEqual(ast.action.right.value, "present")
        self.assertEqual(type(ast.child), Iff)
        self.assertEqual(type(ast.child.left), EaF)
        self.assertEqual(type(ast.child.left.action), Atom)
        self.assertEqual(ast.child.left.action.value, "true")
        self.assertEqual(type(ast.child.left.child), And)
        self.assertEqual(type(ast.child.left.child.left), Atom)
        self.assertEqual(ast.child.left.child.left.value, "future is now")
        self.assertEqual(type(ast.child.left.child.right), Atom)
        self.assertEqual(ast.child.left.child.right.value, "later")
        self.assertEqual(type(ast.child.right), AaU)
        self.assertEqual(type(ast.child.right.action), Atom)
        self.assertEqual(ast.child.right.action.value, "min")
        self.assertEqual(type(ast.child.right.left), Atom)
        self.assertEqual(ast.child.right.left.value, "past")
        self.assertEqual(type(ast.child.right.right), Implies)
        self.assertEqual(type(ast.child.right.right.left), Atom)
        self.assertEqual(ast.child.right.right.left.value, "present")
        self.assertEqual(type(ast.child.right.right.right), Atom)
        self.assertEqual(ast.child.right.right.right.value, "future")